/***********************************************************************

	File:	io.c
	Rev:	a-1
	Date:	02/20/93

	Copyright (c) 1993, 1999 by David M. Warme

************************************************************************

	Routines for Input/Output of point sets and coordinates.
	These routines insulate the rest of the system from the
	exact data type for "coord_t".

	To eliminate errors due to the basic properties of binary
	floating point representation, we scale (i.e., multiply)
	every coordinate by a power of 10 until every coordinate has
	an integral value.  This is done PRIOR to conversion into
	floating point form, and permits us to represent each
	coordinate (as well as delta X and delta Y values) exactly.
	This actually eliminates all numeric errors when computing
	and summing distances in the rectilinear metric (as long as
	the numbers do not exceed the basic precision of a double).

************************************************************************

	Modification Log:

	a-1:	02/20/93	warme
		: Created.

************************************************************************/

#include "steiner.h"


/*
 * Global Routines
 */

void			coord_to_string (char *, coord_t);
void			dist_to_string (char *, dist_t);
struct pset *		get_points (FILE *, int *, coord_t *);
void			init_output_conversion (struct pset *, int, int);
struct numlist *	parse_line_of_numbers (FILE *);
coord_t			read_numlist (struct numlist *, int);


/*
 * Local Routines
 */

static struct numlist *	new_numlist (char *, int);
static int		next_char (FILE *);
static struct numlist *	parse_input_numbers (FILE *);
static struct numlist *	parse_one_number (FILE *, bool);


/*
 * Local Variables
 */

static int		Data_Num_Decimal_Places;
static int		min_precision;

/*
 * This routine reads in a set of points from the standard input.
 */

	struct pset *
get_points (

FILE *			fp,		/* IN - stream to read from */
int *			num_dig,	/* OUT - num coord scale digits */
coord_t *		scale_value	/* OUT - coord scale value */
)
{
int			i;
int			n;
int			max_num_dp;
struct pset *		pts;
struct point *		p;
coord_t			x;
coord_t			y;
struct numlist *	list;
struct numlist *	nlp;

	/* Gather all numbers from the input AS STRINGS.  We also set	*/
	/* the variable "max_num_dp" to be the largest number of	*/
	/* decimal places seen in any of the input numbers.		*/

	list = parse_input_numbers (fp);

	/* Determine coordinate scaling factor. */
	max_num_dp = 0;
	for (nlp = list; nlp NE NULL; nlp = nlp -> next) {
		if (nlp -> num_dp > max_num_dp) {
			max_num_dp = nlp -> num_dp;
		}
	}

	/* Return coordinate scaling info to caller. */
	x = 1.0;
	for (i = 0; i < max_num_dp; i++) {
		x *= 10.0;
	}
	*num_dig = max_num_dp;
	*scale_value = x;

	n = 0;
	for (nlp = list; nlp NE NULL; nlp = nlp -> next) {
		++n;
	}

	if ((n & 1) NE 0) {
		(void) fprintf (stderr, "X coord without Y coord.\n");
		exit (1);
	}
	n >>= 1;

	pts = NEW_PSET (n);

	/* Zero out the entire point set. */
	ZERO_PSET (pts, n);

	pts -> n = n;

	n = 0;
	p = &(pts -> a [0]);

	while (list NE NULL) {
		nlp = list;
		list = list -> next;

		x = read_numlist (nlp, max_num_dp);

		if (list EQ NULL) {
			fatal ("get_points: Bug 1.");
		}

		nlp = list;
		list = list -> next;

		y = read_numlist (nlp, max_num_dp);

		p -> x		= x;
		p -> y		= y;
		p -> pnum	= ((pnum_t) n);

		++n;
		++p;
	}

	if (n NE pts -> n) {
		fatal ("get_points: Bug 2.");
	}

	return (pts);
}

/*
 * This routine "parses" all of the numbers in the input into a list
 * of strings.
 */

	static
	struct numlist *
parse_input_numbers (

FILE *		fp		/* IN - input stream to read from */
)
{
struct numlist *	p;
struct numlist *	root;
struct numlist **	hookp;

	root = NULL;
	hookp = &root;

	for (;;) {
		p = parse_one_number (fp, FALSE);
		if (p EQ NULL) break;

		*hookp = p;
		hookp = &(p -> next);
	}

	return (root);
}

/*
 * Read a single text line of numbers.  We ignore blank lines and
 * lines containing only a comment.  Otherwise, this routine collects
 * numbers in a linked list of numlist structures until the first
 * newline after at least one number is read.
 */

	struct numlist *
parse_line_of_numbers (

FILE *		fp		/* IN - file to read from */
)
{
struct numlist *	root;
struct numlist **	hookp;
struct numlist *	p;

	root = NULL;
	hookp = &root;

	for (;;) {
		p = parse_one_number (fp, TRUE);
		if (p NE NULL) {
			*hookp = p;
			hookp = &(p -> next);
			continue;
		}

		/* We read either EOF or a newline... */
		if (root NE NULL) {
			/* we have one or more numbers -- we don't	*/
			/* really care whether it was newline or EOF!	*/
			break;
		}
		if (feof (fp)) {
			/* It was EOF.  We are not going to get any	*/
			/* more numbers!				*/
			break;
		}
		/* ignore blank/comment line. */
	}

	return (root);
}

/*
 * This routine "parses" the next number from the input stream into
 * a single string list node.
 *
 * It also sets num_dp_ptr to be the number of decimal places occurring
 * in the number parsed.
 */

	static
	struct numlist *
parse_one_number (

FILE *		fp,		/* IN - input stream to read from */
bool		find_newline	/* IN - fail if newline seen before number */
)
{
int		num_dp;
bool		dot_flag;
int		c;
char *		p;
char		buffer [256];

	if (feof (fp)) return (NULL);

	p = &buffer [0];

	do {
		c = next_char (fp);
		if (c < 0) return (NULL);
		if (c EQ '.') break;
		if (c EQ '-') {
			/* Copy the minus sign and proceed. */
			*p++ = c;
			c = next_char (fp);
			break;
		}
		if (ispunct (c)) {
			/* Strip comment line... */
			do {
				c = next_char (fp);
				if (c < 0) return (NULL);
			} while (c NE '\n');
			if (find_newline) return (NULL);
			continue;
		}
		if ((c EQ '\n') AND find_newline) return (NULL);
	} while (NOT isdigit (c));

	/* Strip unnecessary leading zeros. */
	if (c EQ '0') {
		do {
			c = next_char (fp);
		} while (c EQ '0');
		if (c < 0) {
			return (new_numlist ("0", 0));
		}
		else if ((c NE '.') AND (NOT isdigit (c))) {
			ungetc (c, fp);
			return (new_numlist ("0", 0));
		}
	}

	num_dp = 0;
	dot_flag = FALSE;

	for (;;) {
		*p++ = c;
		if (c EQ '.') {
			if (dot_flag) {
				/* OOPS!  More than 1 decimal point	*/
				/* seen in a single number...  Just	*/
				/* terminate this number.		*/
				ungetc (c, fp);
				break;
			}
			dot_flag = TRUE;
		}
		else if (dot_flag) {
			++num_dp;
		}

		c = next_char (fp);
		if (c < 0) break;

		if ((NOT isdigit (c)) AND (c NE '.')) {
			ungetc (c, fp);
			break;
		}
	}

	*p++ = '\0';

	return (new_numlist (&buffer [0], num_dp));
}

/*
 * This routine creates a new string list structure.
 */

	static
	struct numlist *
new_numlist (

char *		s,		/* IN - string to save */
int		num_dp		/* IN - num digits after decimal point */
)
{
struct numlist *	p;
int			n;
size_t			size;

	n = strlen (s) + 1;
	size = offsetof (struct numlist, text [n]);

	p = new (size);

	p -> next = NULL;
	p -> num_dp = num_dp;
	strcpy (p -> text, s);

	return (p);
}

/*
 * This routine converts the coordinate in the given numlist structure
 * to an internal coordinate value, scaled according to the "max_num_dp"
 * parameter.  The given numlist structure is freed.
 */

	coord_t
read_numlist (

struct numlist *	nlp,		/* IN - string list structure. */
int			max_num_dp	/* IN - max # of decimal places. */
)
{
int		c;
char *		p;
coord_t		val;
int		num_dp;
coord_t		sign;
bool		dot_flag;

	p = &(nlp -> text [0]);

	val = 0;
	sign = 1;
	num_dp = 0;
	dot_flag = FALSE;
	if (*p EQ '-') {
		sign = -1;
		++p;
	}

	for (;;) {
		c = *p++;
		if (c EQ '\0') break;
		if (c EQ '.') {
			dot_flag = TRUE;
			continue;
		}

		if (dot_flag) {
			++num_dp;
		}
		val = (val * 10) + (c - '0');
	}

	while (num_dp < max_num_dp) {
		val *= 10;
		++num_dp;
	}

	free ((char *) nlp);

	return (sign * val);
}

/*
 * This routine gets the next character from stdin.  For the case of
 * interactive input, it makes sure that an EOF condition is
 * "permanent" -- that is, getc will not be called again after it
 * indicates EOF.
 */

	static
	int
next_char (

FILE *		fp		/* IN - input stream to read from */
)
{
int		c;

	if (feof (fp)) {
		/* Don't do "getc" again -- EOF is permanent! */
		return (EOF);
	}

	c = getc (fp);

	return (c);
}

/*
 * This routine converts the given internal-form coordinate to a
 * printable ASCII string in external form.  The internal form is an
 * integer (to eliminate numeric problems), but the external data
 * may actually involve decimal fractions.  This routine therefore
 * properly scales the coordinate during conversion to printable form.
 */

	void
coord_to_string (

char *		buf,		/* OUT - buffer to put ASCII text into. */
coord_t		coord		/* IN - coordinate to convert. */
)
{
	/* Just pretend its a distance -- distances certainly	*/
	/* do not have LESS precision than coordinates...	*/
	dist_to_string (buf, coord);
}

/*
 * This routine converts the given internal-form distance to a
 * printable ASCII string in external form.  The internal form is an
 * integer (to eliminate numeric problems), but the external data
 * may actually involve decimal fractions.  This routine therefore
 * properly scales the distance during conversion to printable form.
 */

	void
dist_to_string (

char *		buf,		/* OUT - buffer to put ASCII text into. */
dist_t		dist		/* IN - distance to convert. */
)
{
int		i;
int		digit;
int		mindig;
double		ipart;
double		fpart;
double		div10;
char *		p;
char *		endp;
char		tmp [256];

	if (dist < 0.0) {
		dist = -dist;
		*buf++ = '-';
	}
	if (dist EQ 0) {
		*buf++ = '0';
		*buf++ = '\0';
		return;
	}

	mindig = min_precision;

	memset (tmp, '0', 128);
	p = &tmp [128];
	endp = p;

	ipart = floor (dist);
	while (ipart > 0) {
		div10 = floor (ipart / 10.0);
		digit = ipart - floor (10.0 * div10 + 0.5);
		*--p = digit + '0';
		ipart = div10;
	}
	if ((Data_Num_Decimal_Places EQ 0) AND (mindig EQ 0)) {
		/* The coordinates and edge costs are all integers	*/
		/* (both internally and externally so scaling is a	*/
		/* "no-op").  The metric is not Euclidean, so nothing	*/
		/* irrational appears.  Always output integers, and do	*/
		/* NOT insert a decimal point...			*/

		*endp = '\0';
		strcpy (buf, p);
		return;
	}

	if (p + Data_Num_Decimal_Places > endp) {
		p = endp - Data_Num_Decimal_Places;
	}

	for (i = 0; i <= Data_Num_Decimal_Places; i++) {
		endp [1 - i] = endp [0 - i];
	}
	endp [1 - i] = '.';
	++endp;

	fpart = dist - floor (dist);
	while ((endp - p) < (mindig + 1)) {
		fpart *= 10;
		digit = (int) floor (fpart);
		*endp++ = digit + '0';
		fpart -= ((double) digit);
	}

	*endp = '\0';

	strcpy (buf, p);
}

/*
 * This routine sets up the various parameters needed for outputting
 * scaled coordinates.  We want to print coordinates/distances with
 * the minimum fixed precision whenever this gives the exact result.
 * Otherwise we will print the coordinates/distances with full
 * precision.
 *
 * The minimum fixed precision is exact ONLY when ALL of the following
 * are true:
 *
 *	- The metric must not be EUCLIDEAN, since distances become
 *	  irrational even if the coordinates are finite precision.
 *	  Coordinates of Euclidean Steiner points are also irrational.
 *
 *	- The SCALED vertex coordinates must all be integral.
 *
 *	- The SCALED hyperedge costs must all be integral.
 *
 * Note: we actually check the scaled data for integrality because there
 * are some old FST generators that do not implement scaling!  Such
 * data always contain a scale factor of zero and non-integral coordinates
 * and distances.
 */

	void
init_output_conversion (

struct pset *		pts,		/* IN - list of terminals */
int			metric,		/* IN - the metric */
int			scale		/* IN - num coord scale digits */
)
{
int			i;
struct point *		p;
bool			integral;
double			c;

	Data_Num_Decimal_Places	= scale;

	integral = TRUE;

	if (pts NE NULL) {
		p = &(pts -> a [0]);
		for (i = 0; i < pts -> n; i++, p++) {
			c = floor ((double) (p -> x));
			if (c NE p -> x) {
				integral = FALSE;
				break;
			}
			c = floor ((double) (p -> y));
			if (c NE p -> y) {
				integral = FALSE;
				break;
			}
		}
	}

	if (integral AND (metric NE EUCLIDEAN)) {
		min_precision = 0;
	}
	else {
		min_precision = DBL_DIG + 1;
	}
}
