/***********************************************************************

	File:	sec_comp.c
	Rev:	a-1
	Date:	10/05/96

	Copyright (c) 1996, 1999 by David M. Warme

************************************************************************

	Routines to decompose SEC separation problems into smaller
	and simpler sub-problems.

************************************************************************

	Modification Log:

	a-1:	10/05/96	warme
		: Created.

************************************************************************/

#include "bb.h"
#include "constrnt.h"
#include "sec_comp.h"
#include "sec_heur.h"
#include "steiner.h"


/*
 * Global Routines
 */

struct constraint *	check_component_subtour (bitmap_t *,
						 struct comp *,
						 struct constraint *,
						 double *,
						 bitmap_t *,
						 struct cinfo *);
struct comp *	delete_terminal_from_component (int, struct comp *);
struct comp *	find_congested_components (double *,
					   bitmap_t *,
					   bitmap_t *,
					   bool,
					   struct cinfo *);
int		find_least_congested_terminal (bitmap_t *, struct comp *);
void		free_congested_component (struct comp *);


/*
 * External References
 */

	/* None */

/*
 * Local Types
 */

struct bc {
	struct comp *	comp;		/* component being split */
	struct comp *	list;		/* output list of BCC's */
	int *		dfs;		/* DFS number of each terminal */
	int *		back;		/* lowest DFS num in component */
	int *		stack;		/* base-address of full-set stack */
	int *		sp;		/* current stack pointer */
	int		counter;	/* DFS number generator */
	bitmap_t *	tmap;		/* scratch buffer for new components */
	bitmap_t *	fset_mask;	/* scratch buffer for new components */
	bitmap_t *	fsets_seen;	/* full sets already pushed */
	int		ntmasks;	/* size of tmap */
	int		nfsmasks;	/* size of fset_mask */
	int		ncomps;		/* number of BCC's split off */
	int		max_stack;	/* size of stack */
};


/*
 * Local Routines
 */

static void		bcc (struct bc *, int);
static void		component_terms_to_real_terms (bitmap_t *,
						       struct comp *,
						       bitmap_t *);
static struct comp *	copy_subcomponent (struct comp *,
					   bitmap_t *,
					   bitmap_t *);
static void		create_masks (struct comp *);
static struct comp *	find_first_component (double *,
					      bitmap_t *,
					      bitmap_t *,
					      struct cinfo *);
static void		merge_chains (struct comp *);
static void		reduce_component_in_place (struct comp *);
static struct comp *	simplify_one_component (struct comp *);
static struct comp *	split_biconnected_components (struct comp *);
static void		split_connected_components (struct comp *);
static void		strip_uncongested_terminals (struct comp *);
static int		vacuous_component (struct comp *);

/*
 * Compute the smallest possible subsets of the problem on which to
 * run the SEC separation routines.  We use the following principles
 * to prune down the terminals and full sets examined:
 *
 *	- Let S be a set of terminals that define a violated SEC.  Let
 *	  t be a terminal in S with delta(t) <= 1.  Then there is a
 *	  subset S' of S that does not contain t, yet still defines a
 *	  violated SEC.  Therefore, all such terminals t can be
 *	  iteratively removed from consideration.  (Such terminals are
 *	  NOT "congested".)
 *
 *	- Let the hypergraph of terminals and full sets be partitioned
 *	  into its connected components C1 = (V1,F1), C2 = (V2,F2), ...,
 *	  Ck = (Vk,Fk).  Let S be any violated SEC.  Then there is at
 *	  least one i in 1..k for which Si (a subset of
 *	  [Vi \intersect S]) is a violated SEC.
 *
 *	- [Same, but with C1, C2, ..., Ck being biconnected components.]
 *
 *	- Let F1 be a full set containing exactly two congested terminals
 *	  s and t.  Let S be an SEC violation containing s and t.  Define
 *	  a new hypergraph C' = (V - {s,t} + {u}, F') where u is a new
 *	  terminal not in V, and F' is derived from F by replacing every
 *	  reference to s or t with u.  [Resulting full sets with fewer
 *	  than 2 distinct terminals are deleted from F'.]  Let S be a
 *	  violation in C.  Then S' = S - {s,t} + ({u} if s in S or t in S)
 *	  defines a violation in C'.
 *
 * These rules may be applied recursively to arrive at a final set of
 * congested components that are each usually quit small.
 */

	struct comp *
find_congested_components (

double *		x,		/* IN - current LP solution */
bitmap_t *		vert_mask,	/* IN - set of valid vertices */
bitmap_t *		edge_mask,	/* IN - set of valid hyperedges */
bool			print_flag,	/* IN - TRUE ==> print info */
struct cinfo *		cip		/* IN - compatibility info */
)
{
int			i;
int			j;
int			k;
int			n;
struct comp *		comp;
struct comp *		p;
bitmap_t *		bp1;

	/* Find the first component by iteratively applying the	*/
	/* delta(t) <= 1 rule...				*/

	comp = find_first_component (x, vert_mask, edge_mask, cip);

	if (comp EQ NULL) return (NULL);

	if (print_flag) {
		printf (" %% initially %d congested terminals:\n",
			comp -> num_terms);
#if 0
		bp1 = comp -> tmasks;
		k = 0;
		for (i = 0; i < comp -> num_terms; i++) {
			for (j = 0; j < cip -> num_terms; j++) {
				if (NOT BITON (bp1, j)) continue;
				if (k EQ 0) {
					printf (" %%	");
				}
				printf (" %3d", j);
				++k;
				if (k >= 16) {
					printf ("\n");
					k = 0;
				}
			}
			bp1 += comp -> kmasks;
		}
		if (k > 0) {
			printf ("\n");
		}
#endif
	}

	comp = simplify_one_component (comp);

	if (print_flag) {

		n = 0;
		for (p = comp; p NE NULL; p = p -> next) {
			++n;
		}
		printf (" %% find_congested_components found %d components:\n",
			n);
		n = 0;
		for (p = comp; p NE NULL; p = p -> next) {
			printf (" %%\tcomponent %d:\t%d terms,\t%d fsets\n",
				n++, p -> num_terms, p -> num_fsets);
#if 0
			printf (" %%\t   ");
			for (i = 0; i < p -> num_terms; i++) {
				bp1 = &(p -> tmasks [i * p -> kmasks]);
				k = 0;
				for (j = 0; j < cip -> num_terms; j++) {
					if (NOT BITON (bp1, j)) continue;
					++k;
				}
				if (k NE 1) {
					printf (" {");
				}
				for (j = 0; j < cip -> num_terms; j++) {
					if (NOT BITON (bp1, j)) continue;
					printf (" %d", j);
				}
				if (k NE 1) {
					printf ("}");
				}
			}
			printf ("\n");
#endif
		}
	}

	return (comp);
}

/*
 * This routine finds the initial component by iteratively applying
 * the delta(t) <= 1 rule.  This normally shrinks the problem down
 * substantially.  We then convert the problem into "struct comp"
 * form -- after which we no longer need to refer to the cinfo stuff.
 */

	static
	struct comp *
find_first_component (

double *		x,		/* IN - current LP solution */
bitmap_t *		vert_mask,	/* IN - set of valid vertices */
bitmap_t *		edge_mask,	/* IN - set of valid hyperedges */
struct cinfo *		cip		/* IN - compatibility info */
)
{
int			i;
int			j;
int			k;
int			fs;
int			t;
int			nedges;
int			nverts;
int			nmasks;
int			kmasks;
int			fscount;
int			tcount;
int *			vp1;
int *			vp2;
int *			vp3;
int *			ep1;
int *			ep2;
int *			ep3;
double			sum;
bool			changed;
int *			fset_edge_count;
int *			sp;
int *			new_tnum;
int *			new_fsnum;
struct comp *		newp;
bitmap_t *		bp1;
bitmap_t *		bp2;
bitmap_t *		bp3;
bitmap_t *		bp4;
bitmap_t *		terms_stacked;
bitmap_t *		ctmask;
int *			stack;
double *		tweight;

	nedges = cip -> num_edges;
	nverts = cip -> num_verts;
	nmasks = cip -> num_edge_masks;
	kmasks = cip -> num_vert_masks;

	terms_stacked = NEWA (2 * kmasks, bitmap_t);
	ctmask	      = terms_stacked + kmasks;

	for (i = 0; i < kmasks; i++) {
		terms_stacked [i] = 0;
	}

	/* Compute an "edge count" for each full set. */
	fset_edge_count = NEWA (nedges, int);
	fscount = 0;
	for (i = 0; i < nedges; i++) {
		fset_edge_count [i] = 0;
		if (BITON (edge_mask, i) AND (x [i] > FUZZ)) {
			k = 0;
			vp1 = cip -> edge [i];
			vp2 = cip -> edge [i + 1];
			while (vp1 < vp2) {
				j = *vp1++;
				if (BITON (vert_mask, j)) {
					++k;
				}
			}
			if (k >= 2) {
				fset_edge_count [i] = k - 1;
				++fscount;
			}
		}
	}

	/* Compute the total weight of each terminal... */
	stack = NEWA (nverts, int);
	tweight = NEWA (nverts, double);
	sp = &stack [0];
	tcount = 0;
	for (i = 0; i < nverts; i++) {
		sum = 0.0;
		if (BITON (vert_mask, i)) {
			++tcount;
			ep1 = cip -> term_trees [i];
			ep2 = cip -> term_trees [i + 1];
			while (ep1 < ep2) {
				fs = *ep1++;
				if (fset_edge_count [fs] > 0) {
					sum += x [fs];
				}
			}
		}
		tweight [i] = sum;
	}

	/* Add every terminal with weight <= 1 to a list of terminals	*/
	/* to be deleted...						*/
	for (i = 0; i < nverts; i++) {
		if (NOT BITON (vert_mask, i)) {
			/* Pretend this terminal has already been	*/
			/* stacked and deleted (already has weight 0).	*/
			SETBIT (terms_stacked, i);
		}
		else if (tweight [i] <= (1.0 + FUZZ)) {
			/* Schedule this terminal for deletion! */
			*sp++ = i;
			SETBIT (terms_stacked, i);
		}
	}

	/* Iteratively pop and delete terminals until stack is empty.	*/
	while (sp > stack) {
		t = *--sp;

		/* Prune terminal "t" from the remaining structure... */
		--tcount;
		tweight [t] = 0.0;
		/* Drop one edge from every remaining full set that	*/
		/* contains terminal "t"...				*/
		ep1 = cip -> term_trees [t];
		ep2 = cip -> term_trees [t + 1];
		while (ep1 < ep2) {
			fs = *ep1++;
			if (fset_edge_count [fs] <= 0) continue;
			--(fset_edge_count [fs]);
			if (fset_edge_count [fs] > 0) continue;
			/* Last edge for this full set has now gone	*/
			/* away.  Find the full set's remaining		*/
			/* terminal and subtract this full set's weight	*/
			/* from it.					*/
			--fscount;
			vp1 = cip -> edge [fs];
			vp2 = cip -> edge [fs + 1];
			for (;;) {
				if (vp1 >= vp2) {
					fatal ("find_first_component: Bug 1.");
				}
				j = *vp1++;
				if (tweight [j] > FUZZ) break;
			}
			tweight [j] -= x [fs];

			if ((tweight [j] <= 1.0 + FUZZ) AND
			    (NOT BITON (terms_stacked, j))) {
				/* Schedule this terminal for deletion... */
				*sp++ = j;
				SETBIT (terms_stacked, j);
			}
		}
	}

	/* Construct a mask of the terminals that are left (the	*/
	/* congested terminals).				*/
	bp1 = ctmask;
	bp2 = bp1 + kmasks;
	bp3 = vert_mask;
	bp4 = terms_stacked;
	while (bp1 < bp2) {
		*bp1++ = (*bp3++ & ~(*bp4++));
	}

	if ((tcount EQ 0) OR (fscount EQ 0)) {
		/* Nothing congested left... no components! */
		free ((char *) tweight);
		free ((char *) stack);
		free ((char *) fset_edge_count);
		free ((char *) terms_stacked);
		return (NULL);
	}

	/* Now construct the component for what's left... */

	newp = NEW (struct comp);

	newp -> next		= NULL;
	newp -> num_terms	= tcount;
	newp -> num_fsets	= fscount;
	newp -> flags		= CFLG_CONG;
	newp -> fsnum		= NEWA (fscount, int);
	newp -> x		= NEWA (fscount, double);
	newp -> fsterms		= NEWA (fscount + 1, int *);
	newp -> tfsets		= NEWA (tcount + 1, int *);
	newp -> tviol		= NEWA (tcount, double);
	newp -> tmasks		= NEWA (tcount * kmasks, bitmap_t);
	newp -> kmasks		= kmasks;
	newp -> tmap		= NULL;
	newp -> fset_mask	= NULL;
	newp -> cp		= NULL;

	k = tcount * kmasks;
	for (i = 0; i < k; i++) {
		newp -> tmasks [i] = 0;
	}

	new_tnum = NEWA (nverts, int);
	for (i = 0; i < nverts; i++) {
		new_tnum [i] = -1;
	}
	k = 0;
	bp1 = newp -> tmasks;
	for (i = 0; i < nverts; i++) {
		if (NOT BITON (ctmask, i)) continue;
		/* This terminal was retained... */
		SETBIT (bp1, i);
		new_tnum [i] = k;
		newp -> tviol [k] = 0.0;
		++k;
		bp1 += kmasks;
	}
	if (k NE tcount) {
		fatal ("find_first_component: Bug 2.");
	}

	new_fsnum = NEWA (nedges, int);
	for (i = 0; i < nedges; i++) {
		new_fsnum [i] = -1;
	}

	/* Construct full set info, including fset -> terminals map. */
	j = 0;
	k = 0;
	for (i = 0; i < nedges; i++) {
		if (fset_edge_count [i] < 1) continue;
		new_fsnum [i] = j;
		newp -> fsnum [j] = i;
		newp -> x [j]	= x [i];
		k += (1 + fset_edge_count [i]);
		++j;
	}
	if (j NE fscount) {
		fatal ("find_first_component: Bug 3.");
	}
	vp3 = NEWA (k, int);
	ep3 = NEWA (k, int);
	for (j = 0; j < fscount; j++) {
		newp -> fsterms [j] = vp3;
		i = newp -> fsnum [j];
		vp1 = cip -> edge [i];
		vp2 = cip -> edge [i + 1];
		while (vp1 < vp2) {
			t = *vp1++;
			if (NOT BITON (ctmask, t)) continue;
			t = new_tnum [t];
			if (t < 0) {
				fatal ("find_first_component: Bug 4.");
			}
			*vp3++ = t;
		}
	}
	newp -> fsterms [fscount] = vp3;
	if (vp3 NE newp -> fsterms [0] + k) {
		fatal ("find_first_component: Bug 5.");
	}
	for (i = 0; i < nverts; i++) {
		j = new_tnum [i];
		if (j < 0) continue;
		newp -> tfsets [j] = ep3;
		ep1 = cip -> term_trees [i];
		ep2 = cip -> term_trees [i + 1];
		while (ep1 < ep2) {
			fs = new_fsnum[*ep1++];
			if (fs >= 0) {
				*ep3++ = fs;
			}
		}
	}
	newp -> tfsets [tcount] = ep3;
	if (ep3 NE newp -> tfsets [0] + k) {
		fatal ("find_first_component: Bug 6.");
	}

	free ((char *) new_fsnum);
	free ((char *) new_tnum);
	free ((char *) tweight);
	free ((char *) stack);
	free ((char *) fset_edge_count);
	free ((char *) terms_stacked);

	return (newp);
}

/*
 * This routine simplifies a SINGLE COMPONENT, producing 0 or more
 * simplified components that are concatenated onto the front of the
 * successors that may follow the given component in the linked list.
 */

	static
	struct comp *
simplify_one_component (

struct comp *		comp		/* IN - component to simplify */
)
{
struct comp *		rest;
struct comp *		p;
struct comp *		temp;
struct comp *		done;
struct comp **		done_hookp;

	if (comp EQ NULL) {
		fatal ("simplify_one_component: Bug 1.");
	}

	/* Disconnect this component from any others after it... */
	rest = comp -> next;

	p = comp;		/* one-and-only pending node... */
	p -> next  = NULL;
	done	   = NULL;	/* nothing completed yet... */
	done_hookp = &done;

	while (p NE NULL) {
		/* Process next pending component... */

		/* Create terminal and full-set masks if necessary... */
		create_masks (p);

		if ((p -> flags & CFLG_CONG) EQ 0) {
			/* Reduce down to congested minimum by removing	*/
			/* all uncongested terminals...			*/
			strip_uncongested_terminals (p);
		}
		else if ((p -> flags & CFLG_CC) EQ 0) {
			/* Break off connected components... */
			split_connected_components (p);
		}
		else if ((p -> flags & CFLG_BCC) EQ 0) {
			/* Break off biconnected components... */
			 p = split_biconnected_components (p);
		}
		else if ((p -> flags & CFLG_CHAIN) EQ 0) {
			/* Shrink down long chains of two-terminal	*/
			/* full sets of weight 1.			*/
			merge_chains (p);
		}
		else {
			/* Component is now fully simplified.	*/
			/* Reduce/renumber it...		*/
			reduce_component_in_place (p);
			if (vacuous_component (p)) {
				/* Component disolved into nothing! */
				temp = p -> next;
				free_congested_component (p);
				p = temp;
			}
			else {
				/* Component is DONE.  Move it off	*/
				/* to the "done" list...		*/
				*done_hookp = p;
				done_hookp = &(p -> next);
				p = p -> next;
				*done_hookp = NULL;
			}
		}
	}

	/* Concatenate the completed nodes onto the rest we started with. */
	*done_hookp = rest;

	return (done);
}

/*
 * This routine iteratively strips away all terminals t with delta(t) <= 1
 * until we are left with a "core" of congested terminals.
 */

	static
	void
strip_uncongested_terminals (

struct comp *		comp		/* IN - component to strip */
)
{
int			i;
int			j;
int			k;
int			fs;
int			t;
int			nedges;
int			nterms;
int *			ip1;
int *			ip2;
int *			ip3;
int *			ip4;
double			sum;
int *			fset_term_count;
int *			sp;
bitmap_t *		terms_stacked;
int *			stack;
double *		tweight;

	nterms = comp -> num_terms;
	nedges = comp -> num_fsets;

	k = BMAP_ELTS (nterms);
	terms_stacked = NEWA (k, bitmap_t);

	for (i = 0; i < k; i++) {
		terms_stacked [i] = 0;
	}

	/* Compute a count of valid terminals for each full set. */
	fset_term_count = NEWA (nedges, int);
	for (i = 0; i < nedges; i++) {
		fset_term_count [i] = 0;
		if (NOT BITON (comp -> fset_mask, i)) continue;

		if (comp -> x [i] <= FUZZ) {
			/* Full set not really present... */
			CLRBIT (comp -> fset_mask, i);
			continue;
		}
		ip1 = comp -> fsterms [i];
		ip2 = comp -> fsterms [i + 1];
		k = 0;
		while (ip1 < ip2) {
			j = *ip1++;
			if (BITON (comp -> tmap, j)) {
				++k;
			}
		}
		if (k < 2) {
			/* Full set no longer contributes edges... */
			CLRBIT (comp -> fset_mask, i);
			continue;
		}
		fset_term_count [i] = k;
	}

	/* Compute the total weight of each terminal... */
	tweight = NEWA (nterms, double);
	for (i = 0; i < nterms; i++) {
		/* Start with any weight inherent in the terminal itself... */
		sum = comp -> tviol [i];
		if (BITON (comp -> tmap, i)) {
			ip1 = comp -> tfsets [i];
			ip2 = comp -> tfsets [i + 1];
			while (ip1 < ip2) {
				fs = *ip1++;
				/* Include weight only from full sets	*/
				/* with at least one other valid	*/
				/* terminal...				*/
				if (fset_term_count [fs] < 2) continue;
				sum += comp -> x [fs];
			}
		}
		tweight [i] = sum;
	}

	/* Add every terminal with weight <= 1 to a list of terminals	*/
	/* to be deleted...						*/
	stack = NEWA (nterms, int);
	sp = &stack [0];
	for (i = 0; i < nterms; i++) {
		if (NOT BITON (comp -> tmap, i)) {
			/* Pretend this terminal has already been	*/
			/* stacked and deleted (already has weight 0).	*/
			SETBIT (terms_stacked, i);
		}
		else if (tweight [i] <= (1.0 + FUZZ)) {
			/* Schedule this terminal for deletion! */
			*sp++ = i;
			SETBIT (terms_stacked, i);
		}
	}

	/* Iteratively pop and delete terminals until stack is empty.	*/
	while (sp > stack) {
		t = *--sp;

		/* Prune terminal "t" from the remaining structure... */
		CLRBIT (comp -> tmap, t);
		tweight [t] = 0.0;

		/* Drop one terminal from every remaining full set that	*/
		/* contains terminal "t"...				*/
		ip1 = comp -> tfsets [t];
		ip2 = comp -> tfsets [t + 1];
		while (ip1 < ip2) {
			fs = *ip1++;
			if (fset_term_count [fs] < 2) continue;
			--(fset_term_count [fs]);
			if (fset_term_count [fs] >= 2) continue;
			/* This full set now has fewer than 2 valid	*/
			/* terminals left.  The full set must now go	*/
			/* away -- but first we must find its last	*/
			/* valid terminal and subtract our weight from	*/
			/* it...					*/
			CLRBIT (comp -> fset_mask, fs);
			ip3 = comp -> fsterms [fs];
			ip4 = comp -> fsterms [fs + 1];
			for (;;) {
				if (ip3 >= ip4) {
					fatal ("strip_uncongested_terminals: Bug 1.");
				}
				j = *ip3++;
				if (BITON (comp -> tmap, j)) break;
			}
			tweight [j] -= comp -> x [fs];

			if ((tweight [j] <= 1.0 + FUZZ) AND
			    (NOT BITON (terms_stacked, j))) {
				/* Schedule this terminal for deletion... */
				*sp++ = j;
				SETBIT (terms_stacked, j);
			}
		}
	}

	free ((char *) stack);
	free ((char *) tweight);
	free ((char *) fset_term_count);
	free ((char *) terms_stacked);

	/* This component now contains only congested terminals... */
	comp -> flags |= CFLG_CONG;
}

/*
 * This routine determines the number of connected components within the
 * given component.  If there is only one, it is marked as connected.
 * Otherwise, one connected component is split off from this one.
 */

	static
	void
split_connected_components (

struct comp *		comp		/* IN - component to split */
)
{
int			i;
int			k;
int			t;
int			fs;
int			nterms;
int			nedges;
int			num_term_masks;
int			num_fset_masks;
int *			sp;
int *			ip1;
int *			ip2;
int *			ip3;
int *			ip4;
bitmap_t *		cc_fsets;
struct comp *		p2;
bitmap_t *		cc_tmap;
int *			stack;

	/* Get the masks, just in case... */
	create_masks (comp);

	nterms		= comp -> num_terms;
	nedges		= comp -> num_fsets;

	num_term_masks	= BMAP_ELTS (nterms);
	num_fset_masks	= BMAP_ELTS (nedges);

	/* Make masks of terminals and full sets in the connected	*/
	/* component...							*/
	cc_tmap = NEWA (num_term_masks, bitmap_t);
	for (i = 0; i < num_term_masks; i++) {
		cc_tmap [i] = 0;
	}
	cc_fsets = NEWA (num_fset_masks, bitmap_t);
	for (i = 0; i < num_fset_masks; i++) {
		cc_fsets [i] = 0;
	}

	/* Find a terminal to include in the connected component... */
	for (i = 0; ; i++) {
		if (i >= nterms) {
			/* This component has no terminals!  There is	*/
			/* definitely NOT more than one connected	*/
			/* component!  Reduce this guy, mark him as a	*/
			/* connected component, and get out.		*/
			reduce_component_in_place (comp);
			comp -> flags |= CFLG_CC;
			free ((char *) cc_fsets);
			free ((char *) cc_tmap);
			return;
		}
		if (BITON (comp -> tmap, i)) break;
	}

	/* This is the first terminal in the component.  Stack	*/
	/* initially contains this one terminal...		*/
	SETBIT (cc_tmap, i);

	stack = NEWA (nterms, int);
	sp = &stack [0];
	*sp++ = i;

	/* Scan all terminals reachable from this one via valid full sets. */
	while (sp > &stack [0]) {
		t = *--sp;
		ip1 = comp -> tfsets [t];
		ip2 = comp -> tfsets [t + 1];
		while (ip1 < ip2) {
			fs = *ip1++;
			if (NOT BITON (comp -> fset_mask, fs)) continue;
			if (BITON (cc_fsets, fs)) continue;
			SETBIT (cc_fsets, fs);
			ip3 = comp -> fsterms [fs];
			ip4 = comp -> fsterms [fs + 1];
			while (ip3 < ip4) {
				i = *ip3++;
				if (NOT BITON (comp -> tmap, i)) continue;
				if (BITON (cc_tmap, i)) continue;
				SETBIT (cc_tmap, i);
				*sp++ = i;
			}
		}
	}

	/* The cc_tmap and cc_fsets masks now define a connected	*/
	/* component.  See if this is the entire component...		*/
	for (i = 0; ; i++) {
		if (i >= num_term_masks) {
			/* This component is fully connected as-is. */
			comp -> flags |= CFLG_CC;
			free ((char *) stack);
			free ((char *) cc_fsets);
			free ((char *) cc_tmap);
			return;
		}
		if (cc_tmap [i] NE comp -> tmap [i]) break;
	}

	/* We have identified one of at least TWO connected components.	*/
	/* Copy off the one we identified, and remove it from the	*/
	/* current one...						*/
	p2 = copy_subcomponent (comp, cc_tmap, cc_fsets);
	p2 -> flags |= CFLG_CC;

	/* Insert new component right after this one... */
	p2 -> next = comp -> next;
	comp -> next = p2;

	/* Remove split-off terminals and full sets from this component. */
	for (i = 0; i < num_term_masks; i++) {
		comp -> tmap [i] &= ~cc_tmap [i];
	}
	for (i = 0; i < num_fset_masks; i++) {
		comp -> fset_mask [i] &= ~cc_fsets [i];
	}

	free ((char *) stack);
	free ((char *) cc_fsets);
	free ((char *) cc_tmap);
}

/*
 * This routine splits the given component into its bi-connected-components,
 * assuming that each full set of size n is replaced with the complete graph
 * Kn.  It is essentially the standard algorithm for BCC, but modified to
 * work on a hypergraph -- hyperedges of degree 3 or more are assumed to be
 * inherently bi-connected.
 */

	static
	struct comp *
split_biconnected_components (

struct comp *		comp		/* IN - component to split up */
)
{
int			i;
int			nterms;
int			nedges;
int			max_stack;
struct comp *		p;
struct bc		bc;

	/* First reduce the component.  This takes care of two issues	*/
	/* at the same time: we won't have to worry about terminal or	*/
	/* full sets masks, and we can allocate smaller stacks when	*/
	/* there are no extraneous things lying.			*/

	reduce_component_in_place (comp);

	nterms	= comp -> num_terms;
	nedges	= comp -> num_fsets;

	if ((nterms <= 2) OR (nedges <= 1)) {
		/* Component is already bi-connected. */
		comp -> flags |= CFLG_BCC;
		return (comp);
	}

#if 0
	printf (" %% split_biconnected_components: nterms=%d, nedges=%d\n",
		nterms, nedges);
#endif

	/* We will consume (and dispose of) this component, and		*/
	/* produce a sequence of other sub-components that we prepend	*/
	/* onto the front of the list in place of this one.		*/
	bc.list		= comp -> next;
	bc.comp		= comp;
	comp -> next = NULL;

	bc.dfs		= NEWA (nterms, int);
	bc.back		= NEWA (nterms, int);

	for (i = 0; i < nterms; i++) {
		bc.dfs  [i] = 0;
		bc.back [i] = 0;
	}

	bc.max_stack	= nedges;
	bc.stack	= NEWA (bc.max_stack, int);
	bc.sp		= bc.stack;
	bc.counter	= 0;

	bc.ntmasks	= BMAP_ELTS (nterms);
	bc.nfsmasks	= BMAP_ELTS (nedges);

	bc.tmap		= NEWA (bc.ntmasks, bitmap_t);
	bc.fset_mask	= NEWA (bc.nfsmasks, bitmap_t);
	bc.fsets_seen	= NEWA (bc.nfsmasks, bitmap_t);
	bc.ncomps	= 0;

	for (i = 0; i < bc.nfsmasks; i++) {
		bc.fsets_seen [i] = 0;
	}

	/* Traverse component starting with terminal 0, splitting off	*/
	/* the bi-connected components as we go...			*/
	bcc (&bc, 0);

	if (bc.ncomps > 1) {
		/* Since we broke stuff apart, it is possible that	*/
		/* some terminals that were previously congested have	*/
		/* now lost this property...  Clear the flag on each	*/
		/* generated BCC so that the congested test gets re-run	*/
		/* on them...						*/
		p = bc.list;
		for (i = 0; i < bc.ncomps; i++) {
			p -> flags &= ~CFLG_CONG;
			p = p -> next;
		}
	}

	free ((char *) bc.fsets_seen);
	free ((char *) bc.fset_mask);
	free ((char *) bc.tmap);
	free ((char *) bc.stack);
	free ((char *) bc.back);
	free ((char *) bc.dfs);

	free_congested_component (comp);

	return (bc.list);
}

/*
 * This is the recursive part of the bi-connected-components algorithm.  It
 * is the standard method, with a few tweaks to work on hypergraphs instead.
 */

	static
	void
bcc (

struct bc *		bcp,		/* IN - global BCC data */
int			v		/* IN - current DFS vertex */
)
{
int			i;
int			fs;
int			fs2;
int			w;
int *			ip1;
int *			ip2;
int *			ip3;
int *			ip4;
int *			ip5;
int *			ip6;
int *			sp_save;
int *			stack_endp;
struct comp *		comp;
struct comp *		newp;

	comp = bcp -> comp;

	if ((v < 0) OR (v >= comp -> num_terms)) {
		fatal ("bcc:  Bug 1.");
	}

	++(bcp -> counter);
	bcp -> dfs  [v] = bcp -> counter;
	bcp -> back [v] = bcp -> counter;
	ip1 = comp -> tfsets [v];
	ip2 = comp -> tfsets [v + 1];
	while (ip1 < ip2) {
		fs = *ip1++;
		if ((fs < 0) OR (fs >= comp -> num_fsets)) {
			fatal ("bcc: Bug 2.");
		}
		stack_endp = &(bcp -> stack [bcp -> max_stack]);
		if (NOT BITON (bcp -> fsets_seen, fs)) {
			/* We haven't seen this full set before.  Push	*/
			/* it onto the stack...				*/
			if ((bcp -> sp < bcp -> stack) OR
			    (bcp -> sp >= stack_endp)) {
				fatal ("bcc: Bug 3.");
			}
			*(bcp -> sp)++ = fs;
			SETBIT (bcp -> fsets_seen, fs);
		}
		sp_save = bcp -> sp;
		/* Re-scan the terminals and process them... */
		ip3 = comp -> fsterms [fs];
		ip4 = comp -> fsterms [fs + 1];
		while (ip3 < ip4) {
			w = *ip3++;
			if ((w < 0) OR (w >= comp -> num_terms)) {
				fatal ("bcc: Bug 4.");
			}
			if (bcp -> dfs [w] EQ 0) {
				bcc (bcp, w);
				if (bcp -> back [w] >= bcp -> dfs [v]) {
					/* We have a new BCC! */
					for (i = 0; i < bcp -> ntmasks; i++) {
						bcp -> tmap [i] = 0;
					}
					for (i = 0; i < bcp -> nfsmasks; i++) {
						bcp -> fset_mask [i] = 0;
					}
#if 0
					printf (" %% bcc: popping full sets");
#endif
					do {
						if (bcp -> sp <= bcp -> stack) {
							fatal ("bcc: Bug 5.");
						}
						fs2 = *--(bcp -> sp);
#if 0
						printf (" %d", fs2);
#endif
						SETBIT (bcp -> fset_mask, fs2);
						ip5 = comp -> fsterms [fs2];
						ip6 = comp -> fsterms [fs2 + 1];
						while (ip5 < ip6) {
							i = *ip5++;
							SETBIT (bcp -> tmap, i);
						}
					} while (fs2 NE fs);
#if 0
					printf ("\n");
					print_mask (" % bcc: cterms =",
						    bcp -> tmap,
						    comp -> num_terms);
#endif
					newp = copy_subcomponent (comp,
								  bcp -> tmap,
								  bcp -> fset_mask);
					newp -> flags |= CFLG_BCC;
					newp -> next = bcp -> list;
					bcp -> list = newp;
					++(bcp -> ncomps);
				}
				else if (bcp -> back [w] < bcp -> back [v]) {
					bcp -> back [v] = bcp -> back [w];
				}
			}
			else if (bcp -> back [w] < bcp -> back [v]) {
				bcp -> back [v] = bcp -> back [w];
			}
		}
			
	}
}

/*
 * This routine looks for long chains of full sets where each link Li in
 * the chain is a group of full sets with the following properties:
 *
 *	- Every full set in Li has exactly two terminals A and B.
 *	- Every full set that includes A and B is in Li.
 *	- The sum of the weights of the full sets in Li is 1.
 *
 * Let L0, L1, L2, ..., Lk be such a chain of maximal length.  Let Li
 * contain terminals i-1 and i for i=1..k.  Whenever k > 1 we merge
 * terminals 1, 2, ..., k-1 together -- leaving us with a shorter chain
 * containing 3 terminals and 2 links.
 */

	static
	void
merge_chains (

struct comp *		comp		/* IN - component to merge chains in */
)
{
int			i;
int			fs1;
int			fs2;
int			fs3;
int			t0;
int			t1;
int			t2;
int			t3;
bitmap_t *		bp1;
bitmap_t *		bp2;
bitmap_t *		bp3;
int			num_merge;

	num_merge = 0;

again:

	reduce_component_in_place (comp);

	for (fs2 = 0; fs2 < comp -> num_fsets; fs2++) {
		/* full set must have weight of 1.0 */
		if (comp -> x [fs2] < 1.0 - FUZZ) continue;
		/* full set must have 2 terminals... */
		if (comp -> fsterms [fs2] + 2 NE comp -> fsterms [fs2 + 1]) continue;
		/* Full set fs2 has terminals t1 and t2... */
		t1 = comp -> fsterms [fs2] [0];
		t2 = comp -> fsterms [fs2] [1];
		/* Make sure t1 is in only 2 full sets... */
		if (comp -> tfsets [t1] + 2 NE comp -> tfsets [t1 + 1]) continue;
		/* Make sure t2 is in only 2 full sets... */
		if (comp -> tfsets [t2] + 2 NE comp -> tfsets [t2 + 1]) continue;
		fs1 = comp -> tfsets [t1] [0];
		if (fs1 EQ fs2) {
			fs1 = comp -> tfsets [t1] [1];
		}
		fs3 = comp -> tfsets [t2] [0];
		if (fs3 EQ fs2) {
			fs3 = comp -> tfsets [t2] [1];
		}
		/* Full sets fs1 and fs3 must have weight 1... */
		if (comp -> x [fs1] < 1.0 - FUZZ) continue;
		if (comp -> x [fs3] < 1.0 - FUZZ) continue;
		/* Full sets fs1 and fs3 must have exactly 2 terminals... */
		if (comp -> fsterms [fs1] + 2 NE comp -> fsterms [fs1 + 1]) continue;
		if (comp -> fsterms [fs3] + 2 NE comp -> fsterms [fs3 + 1]) continue;
		t0 = comp -> fsterms [fs1] [0];
		if (t0 EQ t1) {
			t0 = comp -> fsterms [fs1] [1];
		}
		t3 = comp -> fsterms [fs3] [0];
		if (t3 EQ t2) {
			t3 = comp -> fsterms [fs3] [1];
		}
		if (t0 EQ t2) continue;
		if (t1 EQ t3) continue;
		if (t0 EQ t3) continue;

#if 0
		printf (" %% _merge: %d, %d, %d, %d\n", t0, t1, t2, t3);
#endif

		/* Merge terminals t1 and t2 (thus merging full set	*/
		/* fs2 out of existence).  Update full set fs3 to be	*/
		/* (t1, t3) instead of (t2, t3).			*/
		bp1 = &(comp -> tmasks [t1 * comp -> kmasks]);
		bp2 = bp1 + comp -> kmasks;
		bp3 = &(comp -> tmasks [t2 * comp -> kmasks]);
		while (bp1 < bp2) {
			*bp1++ |= *bp3++;
		}
		create_masks (comp);
		CLRBIT (comp -> tmap, t2);
		CLRBIT (comp -> fset_mask, fs2);
		if (comp -> fsterms [fs3] [0] EQ t2) {
			comp -> fsterms [fs3] [0] = t1;
		}
		else if (comp -> fsterms [fs3] [1] EQ t2) {
			comp -> fsterms [fs3] [1] = t1;
		}
		else {
			fatal ("merge_chains: Bug 1.");
		}
		if (comp -> tfsets [t1] [0] EQ fs2) {
			comp -> tfsets [t1] [0] = fs3;
		}
		else if (comp -> tfsets [t1] [1] EQ fs2) {
			comp -> tfsets [t1] [1] = fs3;
		}
		else {
			fatal ("merge_chains: Bug 2.");
		}
		++num_merge;
		goto again;		/* start all over again... */
	}

	if (num_merge > 0) {
#if 0
		printf (" %% merge_chains: reduced from %d terminals to %d\n",
			comp -> num_terms + num_merge, comp -> num_terms);
#endif
	}

	comp -> flags |= CFLG_CHAIN;
}

/*
 * This routine copies off a specified SUBSET of the given component
 * into a new, freshly created component that is returned.
 */

	static
	struct comp *
copy_subcomponent (

struct comp *		comp,		/* IN - component to copy from */
bitmap_t *		vert_mask,	/* IN - vertices to copy */
bitmap_t *		edge_mask	/* IN - hyperedges to copy */
)
{
int			i;
int			j;
int			k;
int			nterms;
int			nedges;
int			kmasks;
int			src;
int			dst;
int			new_num_terms;
int			new_num_fsets;
struct comp *		newp;
int *			ip1;
int *			ip2;
int *			ip3;
int *			term_renum;
int *			fs_renum;
int *			tlist;
int *			fslist;
bitmap_t *		bp1;
bitmap_t *		bp2;
bitmap_t *		bp3;

	/* Force masks to be present... */
	create_masks (comp);

	nterms = comp -> num_terms;
	nedges = comp -> num_fsets;
	kmasks = comp -> kmasks;

	/* Get arrays used to renumber the terminals and full sets... */
	term_renum	= NEWA (nterms, int);
	fs_renum	= NEWA (nedges, int);

	/* Compute renumberings... */
	j = 0;
	for (i = 0; i < nterms; i++) {
		if (BITON (vert_mask, i)) {
			term_renum [i] = j++;
		}
		else {
			term_renum [i] = -1;
		}
	}
	new_num_terms = j;

	j = 0;
	for (i = 0; i < nedges; i++) {
		if (BITON (edge_mask, i)) {
			fs_renum [i] = j++;
		}
		else {
			fs_renum [i] = -1;
		}
	}
	new_num_fsets = j;

	/* Compute size of new full-set-to-terms list... */
	k = 0;
	for (i = 0; i < nedges; i++) {
		if (NOT BITON (edge_mask, i)) continue;
		ip1 = comp -> fsterms [i];
		ip2 = comp -> fsterms [i + 1];
		while (ip1 < ip2) {
			j = *ip1++;
			if (NOT BITON (vert_mask, j)) continue;
			++k;
		}
	}

	/* Create the new component... */
	newp = NEW (struct comp);

	tlist  = NEWA (k, int);
	fslist = NEWA (k, int);

	newp -> next		= NULL;
	newp -> num_terms	= new_num_terms;
	newp -> num_fsets	= new_num_fsets;
	newp -> flags		= comp -> flags;
	newp -> fsnum		= NEWA (new_num_fsets, int);
	newp -> x		= NEWA (new_num_fsets, double);
	newp -> fsterms		= NEWA (new_num_fsets + 1, int *);
	newp -> tfsets		= NEWA (new_num_terms + 1, int *);
	newp -> tviol		= NEWA (new_num_terms, double);
	newp -> tmasks		= NEWA (new_num_terms * kmasks, bitmap_t);
	newp -> kmasks		= kmasks;
	newp -> tmap		= NULL;
	newp -> fset_mask	= NULL;
	newp -> cp		= NULL;

	/* Remap/delete entries in the full-set-to-terminals mapping... */
	dst = 0;
	ip1 = tlist;
	for (src = 0; src < nedges; src++) {
		if (fs_renum [src] < 0) continue;
		ip2 = comp -> fsterms [src];
		ip3 = comp -> fsterms [src + 1];
		newp -> fsterms [dst] = ip1;
		while (ip2 < ip3) {
			j = *ip2++;
			j = term_renum [j];
			if (j >= 0) {
				*ip1++ = j;
			}
		}
		newp -> x [dst] = comp -> x [src];
		newp -> fsnum [dst] = comp -> fsnum [src];
		++dst;
	}
	newp -> fsterms [dst] = ip1;
	if (dst NE new_num_fsets) {
		fatal ("copy_subcomponent: Bug 1.");
	}

	/* Remap/delete entries in the terminal-to-full-sets mapping... */
	dst = 0;
	ip1 = fslist;
	for (src = 0; src < nterms; src++) {
		if (term_renum [src] < 0) continue;
		ip2 = comp -> tfsets [src];
		ip3 = comp -> tfsets [src + 1];
		newp -> tfsets [dst] = ip1;
		while (ip2 < ip3) {
			j = *ip2++;
			j = fs_renum [j];
			if (j >= 0) {
				*ip1++ = j;
			}
		}
		newp -> tviol [dst] = comp -> tviol [src];
		bp1 = &(newp -> tmasks [dst * kmasks]);
		bp2 = bp1 + kmasks;
		bp3 = &(comp -> tmasks [src * kmasks]);
		while (bp1 < bp2) {
			*bp1++ = *bp3++;
		}
		++dst;
	}
	newp -> tfsets [dst] = ip1;
	if (dst NE new_num_terms) {
		fatal ("copy_subcomponent: Bug 2.");
	}

	free ((char *) fs_renum);
	free ((char *) term_renum);

	return (newp);
}

/*
 * This routine reduces the given component in-place by re-numbering
 * all of the terminals and full-sets so that the ones that are not
 * marked in the terminal and full-set masks are gone.
 *
 * This functions as a kind of garbage-compaction effect, but we do
 * NOT actually re-allocate ourselves into smaller memory...
 */

	static
	void
reduce_component_in_place (

struct comp *		comp		/* IN - component to reduce */
)
{
int			i;
int			j;
int			nterms;
int			nedges;
int			kmasks;
int			src;
int			dst;
int			new_num_terms;
int			new_num_fsets;
int *			ip1;
int *			ip2;
int *			ip3;
int *			term_renum;
int *			fs_renum;
bitmap_t *		bp1;
bitmap_t *		bp2;
bitmap_t *		bp3;

	if ((comp -> tmap EQ NULL) AND (comp -> fset_mask EQ NULL)) {
		/* Already reduced... */
		return;
	}

	create_masks (comp);	/* in case one exists, but not the other... */

	nterms = comp -> num_terms;
	nedges = comp -> num_fsets;
	kmasks = comp -> kmasks;

	if ((nterms <= 0) OR (nedges <= 0)) {
		comp -> num_terms = 0;
		comp -> num_fsets = 0;
		free ((char *) (comp -> tmap));
		free ((char *) (comp -> fset_mask));
		comp -> tmap = NULL;
		comp -> fset_mask = NULL;
		return;
	}

	/* Get arrays used to renumber the terminals and full sets... */
	term_renum	= NEWA (nterms, int);
	fs_renum	= NEWA (nedges, int);

	/* Compute renumberings... */
	j = 0;
	for (i = 0; i < nterms; i++) {
		if (BITON (comp -> tmap, i)) {
			term_renum [i] = j++;
		}
		else {
			term_renum [i] = -1;
		}
	}
	new_num_terms = j;

	j = 0;
	for (i = 0; i < nedges; i++) {
		if (BITON (comp -> fset_mask, i)) {
			fs_renum [i] = j++;
		}
		else {
			fs_renum [i] = -1;
		}
	}
	new_num_fsets = j;

	/* Remap/delete entries in the full-set-to-terminals mapping... */
	dst = 0;
	ip1 = comp -> fsterms [0];
	for (src = 0; src < nedges; src++) {
		if (fs_renum [src] < 0) continue;
		ip2 = comp -> fsterms [src];
		ip3 = comp -> fsterms [src + 1];
		comp -> fsterms [dst] = ip1;
		while (ip2 < ip3) {
			j = *ip2++;
			j = term_renum [j];
			if (j >= 0) {
				*ip1++ = j;
			}
		}
		comp -> x [dst] = comp -> x [src];
		comp -> fsnum [dst] = comp -> fsnum [src];
		++dst;
	}
	comp -> fsterms [dst] = ip1;
	if (dst NE new_num_fsets) {
		fatal ("reduce_component_in_place: Bug 1.");
	}
	comp -> num_fsets = new_num_fsets;

	/* Remap/delete entries in the terminal-to-full-sets mapping... */
	dst = 0;
	ip1 = comp -> tfsets [0];
	for (src = 0; src < nterms; src++) {
		if (term_renum [src] < 0) continue;
		ip2 = comp -> tfsets [src];
		ip3 = comp -> tfsets [src + 1];
		comp -> tfsets [dst] = ip1;
		while (ip2 < ip3) {
			j = *ip2++;
			j = fs_renum [j];
			if (j >= 0) {
				*ip1++ = j;
			}
		}
		comp -> tviol [dst] = comp -> tviol [src];
		bp1 = &(comp -> tmasks [dst * kmasks]);
		bp2 = bp1 + kmasks;
		bp3 = &(comp -> tmasks [src * kmasks]);
		while (bp1 < bp2) {
			*bp1++ = *bp3++;
		}
		++dst;
	}
	comp -> tfsets [dst] = ip1;
	if (dst NE new_num_terms) {
		fatal ("reduce_component_in_place: Bug 2.");
	}

	/* Component is now a different size... */
	comp -> num_terms	= new_num_terms;
	comp -> num_fsets	= new_num_fsets;

	free ((char *) fs_renum);
	free ((char *) term_renum);

	/* Free the valid terminals and full sets masks... */
	free ((char *) (comp -> fset_mask));
	free ((char *) (comp -> tmap));
	comp -> tmap		= NULL;
	comp -> fset_mask	= NULL;
}

/*
 * This routine determines whether or not the given component is
 * "vacuous" -- too small to contain a subtour violation.
 */

	static
	int
vacuous_component (

struct comp *		p		/* IN - component to test */
)
{
	if (p -> num_terms <= 0) return (TRUE);
	if (p -> num_fsets <= 1) return (TRUE);
	if (p -> num_terms EQ 1) {
		/* Special check for exactly 1 terminal... */
		if (p -> tviol [0] > FUZZ) {
			/* We have a single terminal, but it is congested! */
			/* This is a valid component... */
			return (FALSE);
		}
		/* Only 1 terminal -- no congestion -- ignore this component */
		return (TRUE);
	}

	/* We have a component worth looking at for violated SEC's... */

	return (FALSE);
}

/*
 * This routine frees up the given congested component.
 */

	void
free_congested_component (

struct comp *		p	/* IN - component to free */
)
{
	if (p -> tmap NE NULL) {
		free ((char *) (p -> tmap));
	}
	if (p -> fset_mask NE NULL) {
		free ((char *) (p -> fset_mask));
	}
	free ((char *) (p -> tfsets [0]));
	free ((char *) (p -> fsterms [0]));
	free ((char *) (p -> tmasks));
	free ((char *) (p -> tviol));
	free ((char *) (p -> tfsets));
	free ((char *) (p -> fsterms));
	free ((char *) (p -> x));
	free ((char *) (p -> fsnum));
	free ((char *) p);
}

/*
 * This routine deletes a specified terminal from the given congested
 * component and then does all possible further simplifications on what
 * is left.  Note that it is possible that the given component splits
 * into several when this is done...
 *
 * Also note that the given component may be MODIFIED -- even FREED!
 */

	struct comp *
delete_terminal_from_component (

int			t,		/* IN - terminal to delete */
struct comp *		comp		/* IN - component to delete from */
)
{
	if (t >= comp -> num_terms) {
		fatal ("delete_terminal_from_component: Bug 1.");
	}

	/* Create masks if not currently present... */
	create_masks (comp);

	if (NOT BITON (comp -> tmap, t)) {
		/* Deleting a terminal that is not present. */
		fatal ("delete_terminal_from_component: Bug 2.");
	}

	/* Remove the terminal from the bit-mask... */
	CLRBIT (comp -> tmap, t);

	/* Now that we have removed a terminal, we can no	*/
	/* longer be certain that the component is: fully	*/
	/* congested, connected, bi-connected, etc...		*/
	comp -> flags &= ~CFLG_ALL;

	comp = simplify_one_component (comp);

	return (comp);
}

/*
 * This routine adds the terminal and full-sets masks to the given
 * component, if they are NOT already present...
 */

	static
	void
create_masks (

struct comp *		p		/* IN - component to put masks on */
)
{
int		i;
int		n;
int		nmasks;
bitmap_t *	mask;

	if (p -> tmap EQ NULL) {
		/* Sprout a new terminal mask with all terminals present. */
		n = p -> num_terms;
		nmasks = BMAP_ELTS (n);
		if (nmasks < 1) {
			nmasks = 1;
		}
		mask = NEWA (nmasks, bitmap_t);
		for (i = 0; i < nmasks; i++) {
			mask [i] = 0;
		}
		for (i = 0; i < n; i++) {
			SETBIT (mask, i);
		}
		p -> tmap = mask;
	}

	if (p -> fset_mask EQ NULL) {
		/* Sprout a new full-set mask with all full-sets present. */
		n = p -> num_fsets;
		nmasks = BMAP_ELTS (n);
		if (nmasks < 1) {
			nmasks = 1;
		}
		mask = NEWA (nmasks, bitmap_t);
		for (i = 0; i < nmasks; i++) {
			mask [i] = 0;
		}
		for (i = 0; i < n; i++) {
			SETBIT (mask, i);
		}
		p -> fset_mask = mask;
	}
}

/*
 * This routine selects a terminal that is a member of the previous
 * solution S and has minimum weight.
 */

	int
find_least_congested_terminal (

bitmap_t *		S,	/* IN - previous separation solution */
struct comp *		comp	/* IN - congested component */
)
{
int			i;
int			j;
int			nterms;
int			nedges;
int *			ip1;
int *			ip2;
int			min_term;
double			w;
double			min_weight;

	nterms	= comp -> num_terms;
	nedges	= comp -> num_fsets;

	min_weight	= 2.0 * nedges;
	min_term	= -1;

	for (i = 0; i < nterms; i++) {
		if ((S NE NULL) AND (NOT BITON (S, i))) continue;
		w = 0.0;
		ip1 = comp -> tfsets [i];
		ip2 = comp -> tfsets [i + 1];
		while (ip1 < ip2) {
			j = *ip1++;
			w += comp -> x [j];
		}
		if (w < min_weight) {
			min_weight = w;
			min_term   = i;
		}
	}
	if (min_term < 0) {
		/* Least congested terminal not found? */
		fatal ("find_least_congested_terminal: Bug 1.");
	}

	return (min_term);
}

/*
 * This routine produces the proper constraints that are represented
 * by the given subset of the given component.  We try to "clean up"
 * the constraint by taking only the congested portion of the
 * terminals in it.  This can strengthen the constraints considerably.
 */

	struct constraint *
check_component_subtour (

bitmap_t *		S,		/* IN - subtour (set of terminals) */
struct comp *		comp,		/* IN - congested component */
struct constraint *	cp,		/* IN - existing constraints */
double *		x,		/* IN - LP solution vector */
bitmap_t *		edge_mask,	/* IN - set of valid hyperedges */
struct cinfo *		cip		/* IN - compatibility info */
)
{
int			kmasks;
struct comp *		vcomps;
struct comp *		tmp;
struct constraint *	cp2;
bool			print_flag;
bitmap_t *		orig_stour;
bitmap_t *		stour;

	kmasks = cip -> num_vert_masks;

	orig_stour = NEWA (2 * kmasks, bitmap_t);
	stour	   = orig_stour + kmasks;

	/* Construct a bitmask of the REAL terminals comprising	*/
	/* this subtour...					*/
	component_terms_to_real_terms (S, comp, orig_stour);

	/* Now that we have the subtour as "real" terminals,	*/
	/* emit the constraint...				*/
	cp = check_subtour (orig_stour, cp, x, edge_mask, cip);

	/* Take the violated subtour and "clean" it up by computing the	*/
	/* congested components for it.  This will get rid of some of	*/
	/* the cruft, and help to strengthen the generated constraints.	*/

	kmasks = comp -> kmasks;

	print_flag = FALSE;
	vcomps = find_congested_components (x,
					    orig_stour,
					    edge_mask,
					    print_flag,
					    cip);

	while (vcomps NE NULL) {
		component_terms_to_real_terms (vcomps -> tmap, vcomps, stour);

		/* Add this constraint (if not already present...) */
		for (cp2 = cp; ; cp2 = cp2 -> next) {
			if (cp2 EQ NULL) {
				cp = check_subtour (stour,
						    cp,
						    x,
						    edge_mask,
						    cip);
				break;
			}
			if (is_equal (stour, cp2 -> mask, kmasks)) break;
		}

		tmp = vcomps -> next;
		vcomps -> next = NULL;
		free_congested_component (vcomps);
		vcomps = tmp;
	}

	free ((char *) orig_stour);

	return (cp);
}

/*
 * This routine converts a terminal subset within a given component
 * back into the corresponding subset of the original terminals.
 * Specifying a NULL pointer for comp_S is a convenient way to specify
 * that the entire component is to be translated back out.
 */

	static
	void
component_terms_to_real_terms (

bitmap_t *		comp_S,	/* IN - subset of terminals in component */
struct comp *		comp,	/* IN - component those terminals are in */
bitmap_t *		real_S	/* OUT - real set of terminals */
)
{
int			i;
int			k;
int			kmasks;
bitmap_t *		bp1;
bitmap_t *		bp2;
bitmap_t *		bp3;

	kmasks = comp -> kmasks;

	/* Zero out entire set of real terminals... */
	for (i = 0; i < kmasks; i++) {
		real_S [i] = 0;
	}

	/* OR together the tmasks of the chosen component terminals... */
	for (i = 0; i < comp -> num_terms; i++) {
		if ((comp_S NE NULL) AND (NOT BITON (comp_S, i))) continue;
		bp1 = real_S;
		bp2 = bp1 + kmasks;
		bp3 = &(comp -> tmasks [i * kmasks]);
		while (bp1 < bp2) {
			*bp1++ |= *bp3++;
		}
	}
}
