/***********************************************************************

	File:	sec_heur.c
	Rev:	a-1
	Date:	05/13/96

	Copyright (c) 1996, 1999 by David M. Warme

************************************************************************

	Heuristic separation procedure for SEC's (Subtour
	Elimination Constraints).  The heuristic is that we reduce
	the problem from full sets (weighted hypergraphs) down to
	standard weighted graphs by computing an MST for each full
	set.

************************************************************************

	Modification Log:

	a-1:	05/13/96	warme
		: Created.

************************************************************************/

#include "bb.h"
#include "constrnt.h"
#include "flow.h"
#include "sec_comp.h"
#include "sec_heur.h"
#include "steiner.h"


/*
 * Global Routines
 */

struct constraint *	check_subtour (bitmap_t *,
				       struct constraint *,
				       double *,
				       bitmap_t *,
				       struct cinfo *);
struct constraint *	enumerate_all_subtours (struct comp *,
						struct constraint *,
						struct bbnode *);
struct constraint *	find_integer_cycles (double *,
					     bitmap_t *,
					     bitmap_t *,
					     struct constraint *,
					     struct cinfo *);
struct constraint *	find_small_subtours (struct comp *,
					     struct constraint *,
					     struct bbnode *);
bool			is_equal (bitmap_t *, bitmap_t *, int);
struct constraint *	sec_flow_heuristic (struct comp *,
					    double *,
					    bitmap_t *,
					    struct cinfo *,
					    struct constraint *);


/*
 * External References
 */

	/* None */


/*
 * Local Types
 */

struct sec_heur_info {

	/* Data used by the flow solver... */
	struct flow_prob	prob;	/* The network flow formulation */
	struct flow_soln	soln;	/* The network flow solution */
	struct flow_temp	temp;	/* Temporary data structures */

	/* Data used to set the arc capacities and modify the	*/
	/* flow network during SEC separation. */
	int		src_arc1;	/* First source arc */
	int		dst_arc1;	/* First destination arc */
	int **		fs_lists;	/* full sets for each arc */
};



/*
 * Local Routines
 */

static void			build_heuristic_SEC_formulation (
					struct comp *		comp,
					struct cinfo *		cip,
					struct sec_heur_info *	hsecp);
static struct constraint *	check_unique_subtour (bitmap_t *,
						      int,
						      struct constraint *);
static struct pset *		compute_terminal_coords (struct comp *,
							 struct cinfo *);
static struct constraint *	cwalk (int,
				       int,
				       bitmap_t *,
				       bitmap_t *,
				       bitmap_t *,
				       bitmap_t *,
				       bitmap_t *,
				       bitmap_t *,
				       bitmap_t *,
				       int *,
				       int *,
				       int *,
				       struct constraint *,
				       struct cinfo *);
static struct constraint *	enumerate_subtours (int,
						    int,
						    double,
						    bitmap_t *,
						    int *,
						    int *,
						    struct comp *,
						    struct constraint *);
static struct constraint *	find_almost_integer_cycles (int,
							    int *,
							    bitmap_t *,
							    bitmap_t *,
							    bitmap_t *,
							    struct cinfo *);
static void			free_heuristic_SEC_formulation (
					struct sec_heur_info *	hsecp);
static int			full_set_mst (struct comp *,
					      int,
					      struct pset *,
					      struct edge *);
static bool			fwalk (int,
				       int,
				       bitmap_t *,
				       bitmap_t *,
				       struct cinfo *);
static struct constraint *	heuristically_find_violated_secs (
					struct comp *		comp,
					bitmap_t *		fset_mask,
					struct cinfo *		cip,
					struct sec_heur_info *	hsecp,
					double *		x,
					struct constraint *	clist);
static void			set_arc_capacities (struct comp *,
						    struct sec_heur_info *);
static void			sort_edges (int, struct edge *, int *);


/*
 * Local Variables
 */

	/* none */

/*
 * This routine heuristically reduces the weighted hypergraph of full
 * sets to a conventional weighted graph by computing a minimum spanning
 * tree for each full set.  We then use the flow formulation of
 * Padberg ("Trees and Cuts" paper) to find SEC violations in the
 * weighted graph.  This method is heuristic since violations will be
 * discovered or not depending on the particular reduction from
 * hypergraph to graph that is chosen.
 */

	struct constraint *
sec_flow_heuristic (

struct comp *		comp,		/* IN - component to separate. */
double *		x,		/* IN - current LP solution. */
bitmap_t *		fset_mask,	/* IN - subset of full sets. */
struct cinfo *		cip,		/* IN - compatability info. */
struct constraint *	cp		/* IN - existing constraints. */
)
{
struct sec_heur_info	sec_heur_info;

	if (cip -> pts EQ NULL) {
		/* This heuristic needs coordinates for each vertex! */
		return (cp);
	}

	if (comp -> num_terms > 1) {
		/* Build an SEC formulation for this component... */
		build_heuristic_SEC_formulation (comp,
						 cip,
						 &sec_heur_info);

		/* Do the heuristic separation... */
		cp = heuristically_find_violated_secs (comp,
						       fset_mask,
						       cip,
						       &sec_heur_info,
						       x,
						       cp);

		free_heuristic_SEC_formulation (&sec_heur_info);
	}

	return (cp);
}

/*
 * This routine builds the initial formulation for the heuristic
 * separation procedure.
 */

	static
	void
build_heuristic_SEC_formulation (

struct comp *		comp,		/* IN - congested component. */
struct cinfo *		cip,		/* IN - compatibility info. */
struct sec_heur_info *	hsecp		/* OUT - SEC flow formulation. */
)
{
int			i;
int			j;
int			k;
int			n;
int			nterms;
int			ntrees;
int			nedges;
int			total_edges;
int			num_arcs;
int			num_nodes;
int			arc_num;
struct pset *		pts;
struct flow_prob *	prob;
struct full_set *	fsp;
struct edge *		fs_mst_edges;
int *			fs_num;
struct edge *		ep1;
struct edge *		ep2;
struct edge *		ep_endp;
int *			nump;
int *			outlist;
int *			inlist;
int *			fslist;
int *			srcp;
int *			dstp;
int *			fs_ip;
int *			ip;
int *			counts;
int **			ptrs;

	nterms	= comp -> num_terms;
	ntrees	= comp -> num_fsets;

	/* Compute actual X,Y coordinates for each terminal in	*/
	/* the congested component...				*/
	pts = compute_terminal_coords (comp, cip);

	/* Tally up the total number of edges needed to compute	*/
	/* a minimum spanning tree for the specified terminals	*/
	/* of each full set.  This is the sum of (|Fi| - 1) for	*/
	/* all full sets Fi...					*/
	total_edges = (comp -> fsterms [ntrees] - comp -> fsterms [0])
		      - ntrees;

	/* Allocate temporary arrays to hold the edges of the	*/
	/* minimum spanning trees of each full set, together	*/
	/* with the full set number for the edge.		*/
	fs_mst_edges	= NEWA (total_edges, struct edge);
	fs_num		= NEWA (total_edges, int);

	/* Compute a Minimum Spanning Tree for each full set.	*/
	/* Record the edges for each, and the full set number	*/
	/* that the edge came from.				*/
	ep1 = fs_mst_edges;
	nump = fs_num;
	for (i = 0; i < ntrees; i++) {
		nedges = full_set_mst (comp, i, pts, ep1);
		ep1 += nedges;
		for (j = 0; j < nedges; j++) {
			*nump++ = i;
		}
	}

	if ((ep1 NE (fs_mst_edges + total_edges)) OR
	    (nump NE (fs_num + total_edges))) {
		fatal ("build_heuristic_SEC_formulation: Bug 1.");
	}

	free ((char *) pts);

	/* Sort the list of all MST edges into increasing order	*/
	/* (primary key = low endpoint, secondary key = high	*/
	/* endpoint).						*/
	sort_edges (total_edges, fs_mst_edges, fs_num);

	/* Scan through the sorted edges and count the number	*/
	/* of unique (a,b) pairs.				*/
	if (total_edges <= 0) {
		nedges = 0;
	}
	else {
		nedges = 1;
		ep1 = &fs_mst_edges [1];
		for (i = 1; i < total_edges; i++, ep1++) {
			if ((ep1 [-1].p1 NE ep1 [0].p1) OR
			    (ep1 [-1].p2 NE ep1 [0].p2)) {
				++nedges;
			}
		}
	}

	/* Compute the total number of directed arcs in the	*/
	/* flow graph.  Each undirected edge becomes a pair of	*/
	/* head-to-tail directed arcs, plus 2 extra arcs per	*/
	/* terminal -- one from the source node and one to the	*/
	/* sink node.						*/
	num_nodes		= nterms + 2;
	num_arcs		= 2 * nedges + 2 * nterms;

	/* Start filling in the flow problem instance... */
	prob = &(hsecp -> prob);
	prob -> num_nodes	= num_nodes;
	prob -> num_arcs	= num_arcs;

	/* Assign node numbers for the source and sink nodes... */
	prob -> source	= nterms;
	prob -> sink	= nterms + 1;

	/* Now that we know how big the directed flow graph is,	*/
	/* allocate storage for the various data structures...	*/

	prob -> out		= NEWA (num_nodes + 1, int *);
	prob -> in		= NEWA (num_nodes + 1, int *);
	hsecp -> fs_lists	= NEWA (2 * nedges + 1, int *);
	prob -> arc_src		= NEWA (num_arcs, int);
	prob -> arc_dst		= NEWA (num_arcs, int);

	outlist			= NEWA (num_arcs, int);
	inlist			= NEWA (num_arcs, int);
	fslist			= NEWA (2 * total_edges, int);

	prob -> capacity	= NEWA (num_arcs, double);

	/* Loop through all of the undirected edges, generating	*/
	/* the pair of directed arcs for each.  For each arc	*/
	/* generated, generate the list of full sets from which	*/
	/* it draws capacity.					*/
	srcp = prob -> arc_src;
	dstp = prob -> arc_dst;

	ep1	= &fs_mst_edges [0];
	ep_endp	= &fs_mst_edges [total_edges];
	fs_ip	= fs_num;
	arc_num	= 0;
	for (i = 0; i < total_edges;) {
		/* Find number of consecutive "copies" of this edge... */
		ep2 = ep1 + 1;
		n = 1;
		for (;;) {
			if (ep2 >= ep_endp) break;
			if (ep1 -> p1 NE ep2 -> p1) break;
			if (ep1 -> p2 NE ep2 -> p2) break;
			++ep2;
			++n;
		}
		/* Record the A -> B arc... */
		*srcp++ = ep1 -> p1;
		*dstp++ = ep1 -> p2;
		hsecp -> fs_lists [arc_num] = fslist;
		for (j = 0; j < n; j++) {
			*fslist++ = fs_ip [j];
		}
		++arc_num;

		/* Record the B -> A arc... */
		*srcp++ = ep1 -> p2;
		*dstp++ = ep1 -> p1;
		hsecp -> fs_lists [arc_num] = fslist;
		for (j = 0; j < n; j++) {
			*fslist++ = fs_ip [j];
		}
		++arc_num;

		/* Advance to the next undirected edge. */
		ep1	+= n;
		fs_ip	+= n;
		i	+= n;
	}
	/* Terminate list of full sets comprising each arc. */
	hsecp -> fs_lists [arc_num] = fslist;

	free ((char *) fs_num);
	free ((char *) fs_mst_edges);

	/* Record the source -> terminal arcs... */
	hsecp -> src_arc1 = arc_num;
	for (i = 0; i < nterms; i++) {
		*srcp++ = prob -> source;
		*dstp++ = i;
		++arc_num;
	}

	/* Record the terminal -> sink arcs... */
	hsecp -> dst_arc1 = arc_num;
	for (i = 0; i < nterms; i++) {
		*srcp++ = i;
		*dstp++ = prob -> sink;
		++arc_num;
	}

	if (arc_num NE num_arcs) {
		fatal ("build_heuristic_SEC_formulation: Bug 2.");
	}

	/* We have now specified the directed flow graph as a	*/
	/* list of directed arcs.  Time to construct the	*/
	/* adjacency lists -- for each node we build a list of	*/
	/* outgoing and incoming arc numbers.  Do the outgoing	*/
	/* lists first...					*/

	counts	= NEWA (num_nodes, int);
	ptrs	= NEWA (num_nodes, int *);

	for (i = 0; i < num_nodes; i++) {
		counts [i] = 0;
	}
	for (i = 0; i < num_arcs; i++) {
		++(counts [prob -> arc_src [i]]);
	}
	ip = outlist;
	for (i = 0; i < num_nodes; i++) {
		ptrs [i] = ip;
		prob -> out [i] = ip;
		ip += counts [i];
	}
	prob -> out [i] = ip;
	for (i = 0; i < num_arcs; i++) {
		j = prob -> arc_src [i];
		ip = ptrs [j]++;
		*ip = i;
	}

	/* Now do the incoming arc lists... */
	for (i = 0; i < num_nodes; i++) {
		counts [i] = 0;
	}
	for (i = 0; i < num_arcs; i++) {
		++(counts [prob -> arc_dst [i]]);
	}
	ip = inlist;
	for (i = 0; i < num_nodes; i++) {
		ptrs [i] = ip;
		prob -> in [i] = ip;
		ip += counts [i];
	}
	prob -> in [i] = ip;
	for (i = 0; i < num_arcs; i++) {
		k = prob -> arc_dst [i];
		ip = ptrs [k]++;
		*ip = i;
	}

	/* Free temporary memory used to build things... */
	free ((char *) counts);
	free ((char *) ptrs);

	/* Initialize the buffers used to hold flow solutions */
	/* and temporary data... */
	create_flow_solution_data (prob, &(hsecp -> soln));
	create_flow_temp_data (prob, &(hsecp -> temp));
}

/*
 * This routine computes an X,Y coordinate for each terminal in the
 * congested component.  This is tricky since there may be "terminals"
 * in the component that represent the MERGING of numerous "real"
 * terminals from the problem data.  We handle this by computing the
 * AVERAGE x,y coordinate of each real terminal in the component
 * terminal.
 */

	static
	struct pset *
compute_terminal_coords (

struct comp *		comp,		/* IN - congested component */
struct cinfo *		cip		/* IN - compatibility info */
)
{
int			i;
int			j;
int			k;
int			nverts;
int			kmasks;
int			comp_nterms;
struct pset *		pts;
struct point *		p1;
bitmap_t *		real_tmask;
coord_t			x;
coord_t			y;

	comp_nterms = comp -> num_terms;
	kmasks	    = comp -> kmasks;

	nverts = cip -> num_verts;

	pts = (struct pset *) new (PSET_SIZE (comp_nterms));

	pts -> n = comp_nterms;
	real_tmask = &(comp -> tmasks [0]);

	for (i = 0; i < comp_nterms; i++, real_tmask += kmasks) {
		/* Compute the "average" of the X,Y coordinates of each	*/
		/* "real" terminal embedded in this "fake" congested	*/
		/* component terminal.  We probably don't even really	*/
		/* care if these additions overflow -- we are simply	*/
		/* choosing SOME spanning tree for the full set...	*/
		x = 0;
		y = 0;
		k = 0;
		for (j = 0; j < nverts; j++) {
			if (BITON (real_tmask, j)) {
				p1 = &(cip -> pts -> a [j]);
				x += p1 -> x;
				y += p1 -> y;
				++k;
			}
		}
		if (k > 0) {
			pts -> a [i].x = x / k;
			pts -> a [i].y = y / k;
		}
		else {
			pts -> a [i].x = 0;
			pts -> a [i].y = 0;
		}
		pts -> a [i].pnum	= i;
	}

	return (pts);
}

/*
 * This routine computes a Minimum Spanning Tree for the given full set
 * from the given congested component.
 */

	static
	int
full_set_mst (

struct comp *		comp,		/* IN - congested component */
int			fsnum,		/* IN - full set number in component */
struct pset *		pts,		/* IN - X,Y coords of terminals */
struct edge *		elist		/* OUT - list of MST edges */
)
{
int			n;
int			t1;
int			t2;
int			nedges;
int			num_mst_edges;
int			num_mst_terms;
struct point *		p1;
struct point *		p2;
struct edge *		ep;
int *			ip1;
int *			ip2;
int *			endp;
struct edge *		edges;

	ip1  = comp -> fsterms [fsnum];
	endp = comp -> fsterms [fsnum + 1];

	n = endp - ip1;

	n = n * (n - 1) / 2;

	edges = NEWA (n, struct edge);

	/* Generate list of all edges, and compute the length	*/
	/* of each.  Note that we list the endpoints of each	*/
	/* edge in increasing order so that edges may be easily	*/
	/* compared and sorted...				*/
	nedges = 0;
	ep = &edges [0];
	num_mst_terms = endp - ip1;
	while (ip1 < endp) {
		t1 = *ip1++;
		p1 = &(pts -> a [t1]);
		ip2 = ip1;
		while (ip2 < endp) {
			t2 = *ip2++;
			p2 = &(pts -> a [t2]);
#if 1
			/* Rectilinear distance. */
			ep -> len = RDIST (p1, p2);
#else
			/* Euclidean distance. */
			ep -> len = EDIST (p1, p2);
#endif
			if (t1 < t2) {
				ep -> p1 = t1;
				ep -> p2 = t2;
			}
			else {
				ep -> p1 = t2;
				ep -> p2 = t1;
			}
			++ep;
			++nedges;
		}
	}

	if (nedges > 0) {
		num_mst_edges = mst_edge_list (num_mst_terms,
					       nedges,
					       edges,
					       elist);
	}
	else {
		num_mst_edges = 0;
	}

	free ((char *) edges);

	return (num_mst_edges);
}

/*
 * This routine sorts the given list of edges (and corresponding full
 * set numbers) on three keys -- first and second endpoint terminal
 * numbers, and then full set number.  Because p1 < p2 for every edge,
 * this is a well-defined ordering of edges for the undirected graph.
 */

	static
	void
sort_edges (

int			n,	/* IN - number of edges to sort. */
struct edge *		edges,	/* IN/OUT - edges to sort. */
int *			fs_num	/* IN/OUT - full set number for each edge. */
)
{
int		h;
struct edge	etmp;
struct edge *	p1;
struct edge *	p2;
struct edge *	p3;
struct edge *	p4;
struct edge *	endp;
int		itmp;
int *		ip1;
int *		ip2;
int *		ip3;

	endp = &edges [n];

	for (h = 1; h <= n; h = 3*h+1) {
	}

	do {
		h = h / 3;
		p4 = &edges [h];
		p1 = p4;
		ip1 = &fs_num [h];
		while (p1 < endp) {
			etmp = *p1;
			itmp = *ip1;
			p2 = p1;
			ip2 = ip1;
			while (TRUE) {
				p3 = (p2 - h);
				ip3 = (ip2 - h);
				if (p3 -> p1 < etmp.p1) break;
				if (p3 -> p1 EQ etmp.p1) {
					if (p3 -> p2 < etmp.p2) break;
					if (p3 -> p2 EQ etmp.p2) {
						if (*ip3 <= itmp) {
							break;
						}
					}
				}
				/* *p3 > *p2 */
				*p2 = *p3;
				p2 = p3;
				*ip2 = *ip3;
				ip2 = ip3;
				if (p2 < p4) break;
			}
			*p2 = etmp;
			*ip2 = itmp;
			++p1;
			++ip1;
		}
	} while (h > 1);
}

/*
 * This routine heuristically identifies violated Subtour Elimination
 * Constraints by finding max-flows (i.e. min-cuts) in the previously
 * constructed directed graph.  The given LP solution is used to set
 * the arc capacities before doing the max-flow.
 *
 * A list of violated SEC constraints is returned.  This may be NULL,
 * in which case no violations were detected.
 */

	static
	struct constraint *
heuristically_find_violated_secs (

struct comp *		comp,		/* IN - congested component */
bitmap_t *		fset_mask,	/* IN - valid full sets */
struct cinfo *		cip,		/* IN - compatibility info */
struct sec_heur_info *	hsecp,		/* IN - directed flow graph */
double *		x,		/* IN - LP solution to separate */
struct constraint *	clist		/* IN - existing constraints */
)
{
int			i;
int			j;
int			n;
int			kmasks;
int			num_nodes;
int			num_arcs;
int			src_arc1;
int			dst_arc1;
bitmap_t *		bp1;
bitmap_t *		bp2;
bitmap_t *		bp3;
double *		flow;
double *		c;
double			save;
bitmap_t *		stour;

	num_nodes	= hsecp -> prob.num_nodes;
	num_arcs	= hsecp -> prob.num_arcs;
	c		= hsecp -> prob.capacity;

	flow		= hsecp -> soln.flow;

	src_arc1	= hsecp -> src_arc1;
	dst_arc1	= hsecp -> dst_arc1;


	/* First we must set the arc capacities from the LP solution. */
	set_arc_capacities (comp, hsecp);

	/* Run one max-flow problem for each terminal vertex... */

	n = num_nodes - 2;

	kmasks = comp -> kmasks;
	stour  = NEWA (kmasks, bitmap_t);

	for (i = 0; i < n; i++) {
		/* Modify problem to find worst SEC involving	*/
		/* terminal i.  Save capacity (for restore).	*/
		save = c [src_arc1 + i];
		c [src_arc1 + i] = INFINITE_FLOW;

		compute_max_flow (&(hsecp -> prob),
				  &(hsecp -> temp),
				  &(hsecp -> soln));

		/* Construct bitmask of "real" terminals... */
		for (j = 0; j < kmasks; j++) {
			stour [j] = 0;
		}
		for (j = 0; j < n; j++) {
			if (NOT BITON (hsecp -> soln.cut, j)) continue;
			bp1 = stour;
			bp2 = bp1 + kmasks;
			bp3 = &(comp -> tmasks [j * kmasks]);
			while (bp1 < bp2) {
				*bp1++ |= *bp3++;
			}
		}

		/* Generate constraint if a violation is present. */
		clist = check_subtour (stour,
				       clist,
				       x,
				       fset_mask,
				       cip);

		/* Restore the modified capacity... */
		c [src_arc1 + i] = save;

		/* Force node i to the far side of the cut... */
		c [dst_arc1 + i] = INFINITE_FLOW;
	}

	free ((char *) stour);

	return (clist);
}

/*
 * This routine sets the all of the arc capacities from the LP solution.
 * There are 3 kinds of arcs to handle:  the "normal" arcs comprised of
 * flow taken from 1 or more full sets, "source" arcs, and "sink" arcs.
 */

	static
	void
set_arc_capacities (

struct comp *		comp,		/* IN - congested component */
struct sec_heur_info *	hsecp		/* IN - directed flow graph */
)
{
int			i;
int			j;
int			num_nodes;
int			nverts;
int			src_arc1;
int			dst_arc1;
double *		c;
int *			ip1;
int *			ip2;
double			sum;
double *		x;

	num_nodes	= hsecp -> prob.num_nodes;
	nverts		= num_nodes - 2;
	c		= hsecp -> prob.capacity;

	src_arc1	= hsecp -> src_arc1;
	dst_arc1	= hsecp -> dst_arc1;

	x = comp -> x;

	/* All arcs before the first "source" arc are "normal" arcs. */
	for (i = 0; i < src_arc1; i++) {
		ip1 = hsecp -> fs_lists [i];
		ip2 = hsecp -> fs_lists [i + 1];
		sum = 0.0;
		while (ip1 < ip2) {
			/* Total weight of all full sets that contribute */
			/* to this arc... */
			j = *ip1++;
			sum += x [j];
		}
		c [i] = 0.5 * sum;
	}

	/* Now initialize all "source" and "sink" arcs... */
	for (i = 0; i < nverts; i++) {
		ip1 = comp -> tfsets [i];
		ip2 = comp -> tfsets [i + 1];
		sum = 0.0;
		while (ip1 < ip2) {
			j = *ip1++;
			sum += x [j];
		}

		if (sum > 2.0) {
			c [src_arc1 + i] = 0.5 * sum - 1.0;
			c [dst_arc1 + i] = 0.0;
		}
		else {
			c [src_arc1 + i] = 0.0;
			c [dst_arc1 + i] = 1.0 - 0.5 * sum;
		}
	}
}

/*
 * This routine frees up the memory allocated by the given SEC max-flow
 * formulation.
 */

	static
	void
free_heuristic_SEC_formulation (

struct sec_heur_info *	hsecp		/* IN - SEC flow formulation */
)
{
	/* Free up the buffers used to hold flow solutions */
	/* and temporary data... */
	free_flow_temp_data (&(hsecp -> temp));
	free_flow_solution_data (&(hsecp -> soln));

	/* Free up the problem formulation... */
	free ((char *) (hsecp -> prob.out [0]));
	free ((char *) (hsecp -> prob.in [0]));
	free ((char *) (hsecp -> fs_lists [0]));

	free ((char *) (hsecp -> prob.out));
	free ((char *) (hsecp -> prob.in));
	free ((char *) (hsecp -> fs_lists));
	free ((char *) (hsecp -> prob.arc_src));
	free ((char *) (hsecp -> prob.arc_dst));
	free ((char *) (hsecp -> prob.capacity));
}

/*
 * This routine locates any cycles formed by hyperedges hyperedges having
 * weight 1.0 in the solution -- in other words, solid integer cycles.
 *
 * Each time we identify a connected component of solid integral edges,
 * we then look for "almost integer" cycles -- integral except for a single
 * fractional edge.  We do this by checking each fractional edge for two
 * or more vertices in common with the integral connected component.
 */

	struct constraint *
find_integer_cycles (

double *		x,		/* IN - LP solution to check */
bitmap_t *		vert_mask,	/* IN - set of valid vertices */
bitmap_t *		edge_mask,	/* IN - set of valid hyperedges */
struct constraint *	cp,		/* IN - previous constraints */
struct cinfo *		cip		/* IN - compatibility info */
)
{
int			i;
int			j;
int			nverts;
int			ntrees;
int			kmasks;
int			nmasks;
bitmap_t *		integral_edges;
bitmap_t *		edges_left;
bitmap_t *		cc_edges;
bitmap_t *		emark;
int *			stack;
int			num_frac;
int			num_cycles;
int *			frac_edge_nums;
struct constraint *	cp1;
struct constraint *	cp2;
bool			cycles_found;
bitmap_t *		vmark;
bitmap_t *		stour;
bitmap_t *		cc_verts;

	nverts = cip -> num_verts;
	ntrees = cip -> num_edges;
	kmasks = cip -> num_vert_masks;
	nmasks = cip -> num_edge_masks;

	stack			= NEWA (2 * nverts, int);
	frac_edge_nums		= NEWA (ntrees, int);
	integral_edges		= NEWA (nmasks, bitmap_t);
	edges_left		= NEWA (nmasks, bitmap_t);
	cc_edges		= NEWA (nmasks, bitmap_t);
	emark			= NEWA (nmasks, bitmap_t);
	vmark			= NEWA (kmasks, bitmap_t);
	stour			= NEWA (kmasks, bitmap_t);
	cc_verts		= NEWA (kmasks, bitmap_t);

	for (i = 0; i < nmasks; i++) {
		integral_edges [i] = 0;
	}

	num_frac = 0;
	for (i = 0; i < ntrees; i++) {
		if (NOT BITON (edge_mask, i)) continue;
		if (x [i] + FUZZ >= 1.0) {
			/* Edge is present with weight 1.0 */
			SETBIT (integral_edges, i);
		}
		else if (x [i] > FUZZ) {
			/* Edge is fractionally present. */
			frac_edge_nums [num_frac++] = i;
		}
	}

	for (i = 0; i < nmasks; i++) {
		edges_left [i] = integral_edges [i];
		emark [i] = 0;
	}

	for (i = 0; i < kmasks; i++) {
		vmark [i] = 0;
	}

	for (i = 0; i < ntrees; i++) {
		/* Find an edge that has not yet been traversed. */
		if (NOT BITON (edges_left, i)) continue;

		/* Prepare a place to hold the vertices and edges	*/
		/* of this integral connected-component...		*/
		for (j = 0; j < kmasks; j++) {
			cc_verts [j] = 0;
		}
		for (j = 0; j < nmasks; j++) {
			cc_edges [j] = 0;
		}

		/* Walk the component containing edge i, looking	*/
		/* for cycles...					*/
		num_cycles = 0;
		cp1 = NULL;
		cp2 = cwalk (i,
			     -1,
			     integral_edges,
			     edges_left,
			     emark,
			     vmark,
			     stour,
			     cc_verts,
			     cc_edges,
			     stack,
			     stack,
			     &num_cycles,
			     cp1,
			     cip);

#if 0
		print_mask (" % Int CC verts:", cc_verts, nverts);
		print_mask (" % Int CC edges:", cc_edges, ntrees);
#endif

		/* Add any cycles we found onto the main list... */
		cycles_found = FALSE;
		while (cp2 NE NULL) {
			cp1 = cp2 -> next;
			cp2 -> next = cp;
			cp = cp2;
			cp2 = cp1;
			cycles_found = TRUE;
		}

		if (cycles_found) continue;

		/* The current connected-component does not contain any	*/
		/* solid integer cycles.  But now that we know all of	*/
		/* the terminals of this CC we can check for any	*/
		/* fractional edges that have at least two vertices in	*/
		/* common with the CC.  These represent "almost		*/
		/* integer" cycles...  The cycle is unique since the	*/
		/* integer CC is acyclic.				*/
		cp2 = find_almost_integer_cycles (num_frac,
						  frac_edge_nums,
						  cc_verts,
						  cc_edges,
						  stour,
						  cip);

		/* Add any "almost integer" cycles onto main list...	*/
		while (cp2 NE NULL) {
			cp1 = cp2 -> next;
			cp2 -> next = cp;
			cp = cp2;
			cp2 = cp1;
		}
	}

	free ((char *) cc_verts);
	free ((char *) stour);
	free ((char *) vmark);
	free ((char *) emark);
	free ((char *) cc_edges);
	free ((char *) edges_left);
	free ((char *) integral_edges);
	free ((char *) frac_edge_nums);
	free ((char *) stack);

	return (cp);
}

/*
 * This routine recursively walks connected components of hyperedges
 * looking for cycles.  It generates a Subtour Elimination Constraint
 * for every cycle found.
 */

	static
	struct constraint *
cwalk (

int			e,		/* IN - edge number */
int			vertex,		/* IN - vertex by which we entered */
					/*      this hyperedge (or -1) */
bitmap_t *		edge_mask,	/* IN - set of valid hyperedges */
bitmap_t *		edges_left,	/* IN - unvisited hyperedges */
bitmap_t *		emark,		/* IN - edges on stack */
bitmap_t *		vmark,		/* IN - vertices on stack */
bitmap_t *		stour,		/* IN - temporary vertex mask */
bitmap_t *		cc_verts,	/* OUT - vertices of connected comp */
bitmap_t *		cc_edges,	/* OUT - edges of connected comp */
int *			stack,		/* IN - base of vertex stack */
int *			sp,		/* IN - top of vertex stack */
int *			num_cycles,	/* IN/OUT - number of cycles found */
struct constraint *	cp,		/* IN - list of constraints */
struct cinfo *		cip		/* IN - compatibility info */
)
{
int			i;
int			j;
int			v;
int			e2;
int			kmasks;
struct constraint *	tmp;
struct constraint *	newp;
int *			ep1;
int *			ep2;
int *			vp1;
int *			vp2;
int *			vp3;

#define CYCLE_LIMIT	250

	/* Stop nearly-infinite loops on highly cyclic stuff... */
	if (*num_cycles >= CYCLE_LIMIT) return (cp);

	if (BITON (emark, e)) {
		/* We have visited this full set before (higher up on	*/
		/* the stack).  We also know it was not the immediately	*/
		/* preceeding full set.  Therefore we have a cycle.	*/
		/* Read the intervening terminals off the stack...	*/
		kmasks = cip -> num_vert_masks;
		for (i = 0; i < kmasks; i++) {
			stour [i] = 0;
		}
		if (sp < &stack [4]) {
			/* Stack too short for this! */
			fatal ("cwalk: Bug 1.");
		}
		for (;;) {
			if (sp < &stack [2]) {
				/* Cycle full set not found on stack! */
				fatal ("cwalk: Bug 2.");
			}
			sp -= 2;
			SETBIT (stour, sp [1]);
			if (sp [0] EQ e) break;
		}
		/* We should see every cycle twice.  Check if we have	*/
		/* seen this one before...				*/
		tmp = check_unique_subtour (stour, kmasks, cp);
		if (tmp NE cp) {
			++(*num_cycles);
		}
		return (tmp);
	}

	/* This full set does NOT currently appear on the	*/
	/* stack.  Push it onto the stack now...		*/
	sp [0] = e;
	SETBIT (emark, e);
	SETBIT (cc_edges, e);

	/* Mark this full set as being traversed AT LEAST once... */
	CLRBIT (edges_left, e);

	/* Now walk to all OTHER terminals in this full set -- all	*/
	/* except for the one (if any) by which we entered this full	*/
	/* set.								*/
	vp1 = cip -> edge [e];
	vp2 = cip -> edge [e + 1];
	while (vp1 < vp2) {
		v = *vp1++;
		if (v EQ vertex) continue;

		if (BITON (vmark, v)) {
			/* We have visited this terminal before (higher	*/
			/* up on the stack).  We also know it was not	*/
			/* the immediately preceeding terminal.		*/
			/* Therefore we have a cycle.  Read it off the	*/
			/* stack...					*/
			kmasks = cip -> num_vert_masks;
			for (j = 0; j < kmasks; j++) {
				stour [j] = 0;
			}
			SETBIT (stour, v);
			vp3 = sp;
			for (;;) {
				if (vp3 <= stack) {
					/* cycle terminal not found on	*/
					/* stack!			*/
					fatal ("cwalk: Bug 3.");
				}
				vp3 -= 2;
				j = vp3 [1];
				if (j EQ v) break;
				SETBIT (stour, j);
			}

			tmp = check_unique_subtour (stour, kmasks, cp);
			if (tmp NE cp) {
				++(*num_cycles);
			}
			cp = tmp;
			continue;
		}

		/* This terminal does NOT currently appear on the	*/
		/* stack.  Push it onto the stack now...		*/
		sp [1] = v;
		SETBIT (vmark, v);

		/* This terminal is part of the connected component... */
		SETBIT (cc_verts, v);

		/* Traverse every valid full set going out of terminal */
		/* v, (except the one we entered terminal v via)... */
		ep1 = cip -> term_trees [v];
		ep2 = cip -> term_trees [v + 1];
		while (ep1 < ep2) {
			e2 = *ep1++;
			if (NOT BITON (edge_mask, e2)) continue;
			if (e2 EQ e) continue;
			/* Recursively walk subtree looking for cycles... */
			cp = cwalk (e2,
				    v,
				    edge_mask,
				    edges_left,
				    emark,
				    vmark,
				    stour,
				    cc_verts,
				    cc_edges,
				    stack,
				    sp + 2,
				    num_cycles,
				    cp,
				    cip);
			if (*num_cycles >= CYCLE_LIMIT) break;
		}
		CLRBIT (vmark, v);	/* Taking vertex v off the stack */
		if (*num_cycles >= CYCLE_LIMIT) break;
	}

	CLRBIT (emark, e);

	return (cp);
}

/*
 * This routine looks for "almost integer" cycles.  We are given an
 * acyclic connected component consisting of integral full sets.  We are
 * also given a list of all of the fractional full sets.  Each fractional
 * full set that shares two or more terminals in common with the integral
 * connected component therefore represents an SEC violation.  Given that
 * they intersect at terminals A and B there is a unique path from A to B
 * in the connected component.  We use a brute force DFS to recover the
 * actual terminals of the cycle.
 */

	static
	struct constraint *
find_almost_integer_cycles (

int			num_frac,	/* IN - number of fractional fsets */
int *			frac_fsnums,	/* IN - list of fractional fsets */
bitmap_t *		cc_terms,	/* IN - terminals of connected comp */
bitmap_t *		cc_fsets,	/* IN - full-sets of connected comp */
bitmap_t *		stour,		/* IN - scratch terminal map */
struct cinfo *		cip		/* IN - compatibility info */
)
{
int			i;
int			j;
int			k;
int			kmasks;
int			fs;
int			t1;
int			t2;
int *			vp1;
int *			vp2;
int *			vp3;
struct constraint *	cp;
bool			found;

	kmasks = cip -> num_vert_masks;

	cp = NULL;

	while (num_frac > 0) {
		fs = *frac_fsnums++;
		--num_frac;

		/* Find each terminal pair (if any) in the CC... */
		vp1 = cip -> edge [fs];
		vp3 = cip -> edge [fs + 1];
		while (vp1 < vp3) {
			t1 = *vp1++;
			if (NOT BITON (cc_terms, t1)) continue;
			vp2 = vp1;
			while (vp2 < vp3) {
				t2 = *vp2++;
				if (NOT BITON (cc_terms, t2)) continue;

				/* We have a new cycle!  Now identify it... */
				for (k = 0; k < kmasks; k++) {
					stour [k] = 0;
				}
				found = fwalk (t1,
					       t2,
					       cc_fsets,
					       stour,
					       cip);
				if (NOT found) {
					fatal ("find_almost_integer_cycle: Bug 1.");
				}

				/* Could have seen this same cycle before... */
				cp = check_unique_subtour (stour, kmasks, cp);
			}
		}
	}

	return (cp);
}

/*
 * This routine recursively finds a path from terminal t1 to terminal
 * t2 in the given (acyclic) connected component of terminals and
 * full sets.  It returns TRUE when the path has been found.
 */

	static
	bool
fwalk (

int			t1,		/* IN - first terminal */
int			t2,		/* IN - last terminal */
bitmap_t *		cc_fsets,	/* IN - full sets of connected comp */
bitmap_t *		path,		/* OUT - terminals along path */
struct cinfo *		cip		/* IN - compatibility info */
)
{
int			i;
int			fs;
int			t3;
int *			ep1;
int *			ep2;
int *			vp1;
int *			vp2;
bool			found;

	if (BITON (path, t1)) {
		/* We've been here before... */
		return (FALSE);
	}

	SETBIT (path, t1);

	if (t1 EQ t2) return (TRUE);

	ep1 = cip -> term_trees [t1];
	ep2 = cip -> term_trees [t1 + 1];
	while (ep1 < ep2) {
		fs = *ep1++;
		if (NOT BITON (cc_fsets, fs)) continue;

		CLRBIT (cc_fsets, fs);
		found = FALSE;
		vp1 = cip -> edge [fs];
		vp2 = cip -> edge [fs + 1];
		while (vp1 < vp2) {
			t3 = *vp1++;
			found = fwalk (t3, t2, cc_fsets, path, cip);
			if (found) break;
		}
		SETBIT (cc_fsets, fs);
		if (found) return (TRUE);
	}

	CLRBIT (path, t1);

	return (FALSE);
}

/*
 * This routine checks the given subtour to see if it is unique.  If so
 * a new constraint is added to the given list.  Otherwise, the given
 * constraint list is returned unchanged.
 */

	static
	struct constraint *
check_unique_subtour (

bitmap_t *		stour,		/* IN - subtour to check */
int			kmasks,		/* IN - number of terminal masks */
struct constraint *	cp		/* IN - current constraint list */
)
{
struct constraint *	p;
bitmap_t *		bp1;
bitmap_t *		bp2;
bitmap_t *		bp3;

	for (p = cp; p NE NULL; p = p -> next) {
		if (is_equal (stour, p -> mask, kmasks)) {
			/* Already seen this subtour... */
			return (cp);
		}
	}

#if 0
	print_mask (" % Integral cycle:", stour, kmasks * BPW);
#endif

	/* Not seen before -- record it. */
	p = NEW (struct constraint);
	bp1 = NEWA (kmasks, bitmap_t);

	p -> next	= cp;
	p -> iteration	= 0;
	p -> type	= CT_SUBTOUR;
	p -> mask	= bp1;
	bp2 = stour;
	bp3 = bp2 + kmasks;
	while (bp2 < bp3) {
		*bp1++ = *bp2++;
	}
	return (p);
}

/*
 * This routine returns TRUE if-and-only-if the two bit-masks
 * are identical.
 */

	bool
is_equal (

bitmap_t *	bp1,		/* IN - first set. */
bitmap_t *	bp2,		/* IN - second set. */
int		nmasks		/* IN - number of masks in set. */
)
{
int		i;

	for (i = 0; i < nmasks; i++) {
		if (*bp1 NE *bp2) return (FALSE);
		++bp1;
		++bp2;
	}

	return (TRUE);
}

/*
 * This routine finds violated SEC's by exhaustively enumerating all
 * subsets of size 2 or more.  We return a list of the constraints that
 * were found.
 */

	struct constraint *
enumerate_all_subtours (

struct comp *		comp,		/* IN - component to check. */
struct constraint *	cp,		/* IN - existing list of constraints */
struct bbnode *		node		/* IN - current node. */
)
{
int			i;
int			k;
int			nterms;
int			kmasks;
double			rhs;
bitmap_t *		chosen;
int *			clist;

	nterms = comp -> num_terms;

	k = BMAP_ELTS (nterms);

	clist  = NEWA (nterms, int);
	chosen = NEWA (k, bitmap_t);

	for (i = 0; i < k; i++) {
		chosen [i] = 0;
	}

#if 0
	printf (" %% Exhaustively enumerating %d congested terminals.\n", nterms);
#endif


	for (i = 2; i <= nterms; i++) {
		/* Look for subtours of size "i". */
		rhs = ((double) (i - 1)) + FUZZ;
		cp = enumerate_subtours (i,
					 0,
					 rhs,
					 chosen,
					 &clist [0],
					 &clist [0],
					 comp,
					 cp);
	}

	free ((char *) chosen);
	free ((char *) clist);

	return (cp);
}

/*
 * This routine finds violated SEC's by explicitly enumerating subsets
 * of increasing cardinality.  Only the terminals of the given congested
 * component are examined.  We return a list of the constraints that
 * were found.
 */

	struct constraint *
find_small_subtours (

struct comp *		comp,		/* IN - component to check. */
struct constraint *	cp,		/* IN - existing list of constraints */
struct bbnode *		node		/* IN - current node. */
)
{
int			i;
int			k;
int			nterms;
int			ntrees;
int			klimit;
int			ticks_limit;
cpu_time_t		t0;
cpu_time_t		t1;
bitmap_t *		chosen;
int *			clist;
double			est;
double			rhs;
char			tbuf [32];

	ntrees = comp -> num_fsets;
	nterms = comp -> num_terms;

	k = BMAP_ELTS (nterms);

	clist  = NEWA (nterms, int);
	chosen = NEWA (k, bitmap_t);

	for (i = 0; i < k; i++) {
		chosen [i] = 0;
	}

	printf (" %% Enumerating %d congested terminals.\n", nterms);

	t0 = get_cpu_time ();

	if ((node -> br1cnt <= 0) AND (node -> depth <= 4)) {
		/* Root node and first 4 "var=0" branches should */
		/* be done more thoroughly...			 */
		klimit = 2;
		if (nterms <= SEC_ENUM_LIMIT) {
			klimit = nterms;
		}
		else if (nterms <= 2 * SEC_ENUM_LIMIT) {
			klimit = 5;
		}
		else if (nterms <= 3 * SEC_ENUM_LIMIT) {
			klimit = 3;
		}
		ticks_limit = nterms * TICKS_PER_SEC;
	}
	else {
		/* On other nodes, enumerate only pairs... */
		klimit = 2;
		if (nterms <= 2 * SEC_ENUM_LIMIT) {
			klimit = 5;
		}
		else if (nterms <= 3 * SEC_ENUM_LIMIT) {
			klimit = 3;
		}
		ticks_limit = nterms * TICKS_PER_SEC;
	}

	/* Don't need to check 2-terminal subtours if they are	*/
	/* already present in the constraint pool!		*/
	i = Seed_Pool_With_2SECs ? 3 : 2;

	for (; i <= klimit; i++) {
#if 0
		printf (" %% Checking %d subtours\n", i);
#endif

		/* Look for subtours of size "i". */
		rhs = ((double) (i - 1)) + FUZZ;
		cp = enumerate_subtours (i,
					 0,
					 rhs,
					 chosen,
					 &clist [0],
					 &clist [0],
					 comp,
					 cp);

		/* Determine how long this took... */
		t1 = get_cpu_time ();
#if 0
		convert_cpu_time (t1 - t0, tbuf);
		(void) printf (" %%	%s seconds\n", tbuf);
#endif

		/* if we found any violations, get out! */
		if (cp NE NULL) break;

		/* Estimate how long the next pass will take... */
		est = ((double) (nterms - i)) / ((double) (i + 1));
		est *= (t1 - t0);
		if (est > ticks_limit) {
			/* Do not take more than specified	*/
			/* number of ticks...			*/
			break;
		}
		t0 = t1;
	}

	free ((char *) chosen);
	free ((char *) clist);

	return (cp);
}

/*
 * This routine recursively enumerates subsets of size k of the
 * congested terminals.
 */

	static
	struct constraint *
enumerate_subtours (

int			k,		/* IN - size of subset to choose */
int			t,		/* IN - next terminal to take/omit */
double			rhs,		/* IN - required right-hand-side */
bitmap_t *		chosen,		/* IN - component terminals taken */
int *			start,		/* IN - start of chosen terminals */
int *			end,		/* IN - end of chosen terminals */
struct comp *		comp,		/* IN - component to enumerate */
struct constraint *	cp		/* IN - existing constraints */
)
{
int			i;
int			j;
int			nterms;
int			ntrees;
int			kmasks;
int			limit;
int *			ip1;
int *			ip2;
bitmap_t *		bp1;
bitmap_t *		bp2;
struct constraint *	newp;
double			sum;

	nterms = comp -> num_terms;

	if (k <= 0) {
		/* We have chosen a complete subset of the desired	*/
		/* cardinality.  Check it for violations...		*/
		ntrees = comp -> num_fsets;
		kmasks = comp -> kmasks;
		sum = 0.0;
		for (i = 0; i < ntrees; i++) {
			ip1 = comp -> fsterms [i];
			ip2 = comp -> fsterms [i + 1];
			k = -1;
			while (ip1 < ip2) {
				t = *ip1++;
				if (BITON (chosen, t)) {
					++k;
				}
			}
			if (k > 0) {
				sum += (k * (comp -> x [i]));
			}
		}
		ip1 = start;
		while (ip1 < end) {
			t = *ip1++;
			sum += comp -> tviol [t];
		}

		if (sum > rhs) {
			/* We have a violation!  Assemble a mask of the	*/
			/* actual terminals in the subtour from the	*/
			/* chosen terminals of the congested component. */
			bp1 = NEWA (kmasks, bitmap_t);

			for (i = 0; i < kmasks; i++) {
				bp1 [i] = 0;
			}
			ip1 = start;
			while (ip1 < end) {
				i = *ip1++;
				bp2 = &(comp -> tmasks [i * kmasks]);
				for (j = 0; j < kmasks; j++) {
					bp1 [j] |= *bp2++;
				}
			}

			newp = NEW (struct constraint);
			newp -> next		= cp;
			newp -> iteration	= 0;
			newp -> type		= CT_SUBTOUR;
			newp -> mask		= bp1;
			cp = newp;
#if 0
			print_mask (" %% subtour:", bp1, kmasks * BPW);
#endif
		}

		return (cp);
	}

	limit = nterms - k;

	while (t <= limit) {
		/* Try taking the current terminal... */
		SETBIT (chosen, t);
		*end = t;
		cp = enumerate_subtours (k - 1,
					 t + 1,
					 rhs,
					 chosen,
					 start,
					 end + 1,
					 comp,
					 cp);
		CLRBIT (chosen, t);
		++t;
	}

	return (cp);
}

/*
 * This routine checks the given subset of terminals to see if it defines
 * a violated subtour constraint.  If so it adds the constraint to the
 * current list of constraints.
 */

	struct constraint *
check_subtour (

bitmap_t *		stour,		/* IN - subset to check */
struct constraint *	clist,		/* IN - existing constraints */
double *		x,		/* IN - current LP solution */
bitmap_t *		edge_mask,	/* IN - set of valid hyperedges */
struct cinfo *		cip		/* IN - compatibility info */
)
{
int			i;
int			j;
int			kmasks;
int			nedges;
int			ssize;
int			count;
int *			vp1;
int *			vp2;
bitmap_t *		bp1;
bitmap_t		mask;
double			total;
double			coeff;
struct constraint *	cp;

	nedges = cip -> num_edges;
	kmasks = cip -> num_vert_masks;

	/* Get size of subtour - 1... */
	ssize = -1;
	bp1 = stour;
	for (i = 0; i < kmasks; i++) {
		mask = *bp1++;
		ssize += NBITSON (mask);
	}

	if (ssize <= 0) return (clist);

	total = 0.0;
	for (j = 0; j < nedges; j++) {
		if (NOT BITON (edge_mask, j)) continue;
		count = -1;
		vp1 = cip -> edge [j];
		vp2 = cip -> edge [j + 1];
		while (vp1 < vp2) {
			i = *vp1++;
			if (BITON (stour, i)) {
				++count;
			}
		}
		if (count <= 0) continue;
		coeff = ((double) count);
		total += coeff * x [j];
	}

	if (total <= ((double) ssize) + FUZZ) {
		/* Constraint is not violated... */
		return (clist);
	}

	/* Create new constraint and add it to the list... */

	bp1 = NEWA (kmasks, bitmap_t);
	for (i = 0; i < kmasks; i++) {
		bp1 [i] = stour [i];
	}

	cp = NEW (struct constraint);
	cp -> next	= clist;
	cp -> iteration	= 0;
	cp -> type	= CT_SUBTOUR;
	cp -> mask	= bp1;

#if 0
	print_mask (" %% subtour:", stour, cip -> num_terms);
#endif
#if 0
	debug_print_constraint (" % Subtour:  ",
				" %\t",
				cp,
				x,
				edge_mask,
				cip);
#endif

	return (cp);
}
